# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl import Atom, Atoms, Orbital, PeriodicTable

pytestmark = [pytest.mark.atom, pytest.mark.atoms]


@pytest.fixture
def setup():
    class t:
        def __init__(self):
            self.C = Atom["C"]
            self.C3 = Atom("C", [-1] * 3)
            self.Au = Atom("Au")
            self.PT = PeriodicTable()

    return t()


def test_create1(setup):
    atom1 = Atoms([setup.C, setup.C3, setup.Au])
    atom2 = Atoms(["C", "C", "Au"])
    atom3 = Atoms(["C", 6, "Au"])
    atom4 = Atoms(["Au", 6, "C"])
    assert atom2 == atom3
    assert atom2 != atom4
    assert atom2.hassame(atom4)


def test_create2():
    atom = Atoms(Atom(6, R=1.45), na=2)
    atom = Atoms(atom, na=4)
    assert atom[0].maxR() == 1.45
    for ia in range(len(atom)):
        assert atom.maxR(True)[ia] == 1.45


def test_create_map():
    atom = Atoms(map(lambda x: x, [1, 2, 3]))
    assert len(atom) == 3


def test_len(setup):
    atom = Atoms([setup.C, setup.C3, setup.Au])
    assert len(atom) == 3
    assert len(atom.q0) == 3


def test_get1():
    atoms = Atoms(["C", "C", "Au"])
    assert atoms[2] == Atom("Au")
    assert atoms["Au"] == Atom("Au")
    assert atoms[0] == Atom("C")
    assert atoms[1] == Atom("C")
    assert atoms["C"] == Atom("C")
    assert atoms[0:2] == [Atom("C")] * 2
    assert atoms[1:] == [Atom("C"), Atom("Au")]


def test_set1():
    # Add new atoms to the set
    atom = Atoms(["C", "C"])
    assert atom[0] == Atom("C")
    assert atom[1] == Atom("C")
    atom[1] = Atom("Au")
    assert atom[0] == Atom("C")
    assert atom[1] == Atom("Au")
    atom["C"] = Atom("Au")
    assert atom[0] == Atom("Au")
    assert atom[1] == Atom("Au")


@pytest.mark.filterwarnings("ignore", message="*Replacing atom")
def test_set2():
    # Add new atoms to the set
    atom = Atoms(["C", "C"])
    assert atom[0] == Atom("C")
    assert atom[1] == Atom("C")
    assert len(atom.atom) == 1
    atom[1] = Atom("Au", [-1] * 2)
    assert atom[0] == Atom("C")
    assert atom[1] != Atom("Au")
    assert atom[1] == Atom("Au", [-1] * 2)
    assert len(atom.atom) == 2
    atom["C"] = Atom("Au", [-1] * 2)
    assert atom[0] != Atom("Au")
    assert atom[0] == Atom("Au", [-1] * 2)
    assert atom[1] != Atom("Au")
    assert atom[1] == Atom("Au", [-1] * 2)
    assert len(atom.atom) == 1


def test_set3():
    # Add new atoms to the set
    atom = Atoms(["C"] * 10)
    atom[range(1, 4)] = Atom("Au", [-1] * 2)
    assert atom[0] == Atom("C")
    for i in range(1, 4):
        assert atom[i] != Atom("Au")
        assert atom[i] == Atom("Au", [-1] * 2)
    assert atom[4] != Atom("Au")
    assert atom[4] != Atom("Au", [-1] * 2)
    assert len(atom.atom) == 2
    atom[1:4] = Atom("C")
    assert len(atom.atom) == 2


@pytest.mark.filterwarnings("ignore", message="*Replacing atom")
def test_replace1():
    # Add new atoms to the set
    atom = Atoms(["C"] * 10 + ["B"] * 2)
    atom[range(1, 4)] = Atom("Au", [-1] * 2)
    assert atom[0] == Atom("C")
    for i in range(1, 4):
        assert atom[i] != Atom("Au")
        assert atom[i] == Atom("Au", [-1] * 2)
    assert atom[4] != Atom("Au")
    assert atom[4] != Atom("Au", [-1] * 2)
    assert len(atom.atom) == 3
    atom.replace(atom[0], Atom("C", [-1] * 2))
    assert atom[0] == Atom("C", [-1] * 2)
    assert len(atom.atom) == 3
    assert atom[0] == Atom("C", [-1] * 2)
    for i in range(4, 10):
        assert atom[i] == Atom("C", [-1] * 2)
    for i in range(1, 4):
        assert atom[i] == Atom("Au", [-1] * 2)
    for i in range(10, 12):
        assert atom[i] == Atom("B")


@pytest.mark.filterwarnings("ignore", message="*Substituting atom")
@pytest.mark.filterwarnings("ignore", message="*Replacing atom")
def test_replace2():
    # Add new atoms to the set
    atom = Atoms(["C"] * 10 + ["B"] * 2)
    atom.replace(range(1, 4), Atom("Au", [-1] * 2))
    assert atom[0] == Atom("C")
    for i in range(1, 4):
        assert atom[i] != Atom("Au")
        assert atom[i] == Atom("Au", [-1] * 2)
    assert atom[4] != Atom("Au")
    assert atom[4] != Atom("Au", [-1] * 2)
    assert len(atom.atom) == 3

    # Second replace call (equivalent to replace_atom)
    atom.replace(atom[0], Atom("C", [-1] * 2))
    assert atom[0] == Atom("C", [-1] * 2)
    assert len(atom.atom) == 3
    assert atom[0] == Atom("C", [-1] * 2)
    for i in range(4, 10):
        assert atom[i] == Atom("C", [-1] * 2)
    for i in range(1, 4):
        assert atom[i] == Atom("Au", [-1] * 2)
    for i in range(10, 12):
        assert atom[i] == Atom("B")


def test_append1():
    # Add new atoms to the set
    atom1 = Atoms(["C", "C"])
    assert atom1[0] == Atom("C")
    assert atom1[1] == Atom("C")
    atom2 = Atoms([Atom("C", tag="DZ"), Atom[6]])
    assert atom2[0] == Atom("C", tag="DZ")
    assert atom2[1] == Atom("C")

    atom = atom1.append(atom2)
    assert atom[0] == Atom("C")
    assert atom[1] == Atom("C")
    assert atom[2] == Atom("C", tag="DZ")
    assert atom[3] == Atom("C")

    atom = atom1.append(Atom(6, tag="DZ"))
    assert atom[0] == Atom("C")
    assert atom[1] == Atom("C")
    assert atom[2] == Atom("C", tag="DZ")

    atom = atom1.append([Atom(6, tag="DZ"), Atom[6]])
    assert atom[0] == Atom("C")
    assert atom[1] == Atom("C")
    assert atom[2] == Atom("C", tag="DZ")
    assert atom[3] == Atom("C")


def test_compare1():
    # Add new atoms to the set
    atom1 = Atoms([Atom("C", tag="DZ"), Atom[6]])
    atom2 = Atoms([Atom[6], Atom("C", tag="DZ")])
    assert atom1.hassame(atom2)
    assert not atom1.equal(atom2)


def test_in1():
    # Add new atoms to the set
    atom = Atoms(["C", "C"])
    assert Atom[6] in atom
    assert Atom[1] not in atom


def test_iter1():
    # Add new atoms to the set
    atom = Atoms(["C", "C"])
    for a in atom.iter():
        assert a == Atom[6]
    for a, idx in atom.iter(True):
        assert a == Atom[6]
        assert len(idx) == 2

    atom = Atoms(["C", "Au", "C", "Au"])
    for i, aidx in enumerate(atom.iter(True)):
        a, idx = aidx
        if i == 0:
            assert a == Atom[6]
            assert (idx == [0, 2]).all()
        elif i == 1:
            assert a == Atom["Au"]
            assert (idx == [1, 3]).all()
        assert len(idx) == 2


def test_reduce1():
    atom = Atoms(["C", "Au"])
    atom = atom.sub(0)
    atom1 = atom.reduce()
    assert atom[0] == Atom[6]
    assert len(atom) == 1
    assert len(atom.atom) == 2
    assert atom1[0] == Atom[6]
    assert atom1[0] != Atom[8]
    assert len(atom1) == 1
    assert len(atom1.atom) == 1
    atom.reduce(True)
    assert atom[0] == Atom[6]
    assert len(atom) == 1
    assert len(atom.atom) == 1


def test_remove1():
    atom = Atoms(["C", "Au"])
    atom = atom.remove(1)
    atom = atom.reduce()
    assert atom[0] == Atom[6]
    assert len(atom) == 1
    assert len(atom.atom) == 1


def test_reorder1():
    atom = Atoms(["C", "Au"])
    atom = atom.sub(1)
    atom1 = atom.reorder()
    # Check we haven't done anything to the original Atoms object
    assert atom[0] == Atom["Au"]
    assert atom.species[0] == 1
    assert len(atom) == 1
    assert len(atom.atom) == 2
    assert atom1[0] == Atom["Au"]
    assert atom1.species[0] == 0
    assert len(atom1) == 1
    assert len(atom1.atom) == 2
    # Do in-place
    atom.reorder(True)
    assert atom[0] == Atom["Au"]
    assert atom.species[0] == 0
    assert len(atom) == 1
    assert len(atom.atom) == 2


def test_reorder2():
    atom1 = Atoms(["C", "Au"])
    atom2 = atom1.reorder()
    assert atom1 == atom2


def test_charge1():
    atom = Atoms(["C", "Au"])
    assert len(atom.q0) == 2
    assert atom.q0.sum() == pytest.approx(0.0)


def test_charge_diff():
    o1 = Orbital(1.0, 1.0)
    o2 = Orbital(1.0, 0.5)
    a1 = Atom(5, [o1, o2, o1, o2])
    a2 = Atom(5, [o1, o2, o1, o2, o1, o1])
    a = Atoms([a1, a2])
    assert len(a.q0) == 2
    assert a.q0.sum() == pytest.approx(8)
    assert np.allclose(a.q0, [3, 5])


def test_index1():
    atom = Atoms(["C", "Au"])
    with pytest.raises(KeyError):
        atom.index(Atom("B"))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import math as m
from itertools import product

import numpy as np
import pytest

from sisl import (
    BandStructure,
    BrillouinZone,
    Lattice,
    LatticeChild,
    MonkhorstPack,
    SislError,
    geom,
)

pytestmark = [pytest.mark.physics, pytest.mark.brillouinzone, pytest.mark.bz]


@pytest.fixture
def setup():
    class t:
        def __init__(self):
            self.s1 = Lattice(1, nsc=[3, 3, 1])
            self.s2 = Lattice([2, 2, 10, 90, 90, 60], [5, 5, 1])
            self.s3 = Lattice(1, nsc=[3, 1, 3])

    return t()


class TestBrillouinZone:
    def test_bz1(self, setup):
        bz = BrillouinZone(1.0)
        str(bz)
        bz.weight
        bz = BrillouinZone(setup.s1)
        assert len(bz) == 1
        assert np.allclose(bz.tocartesian([0, 0, 0]), [0] * 3)
        assert np.allclose(bz.tocartesian([0.5, 0, 0]), [m.pi, 0, 0])
        assert np.allclose(bz.toreduced([0, 0, 0]), [0] * 3)
        assert np.allclose([0.5, 0, 0], bz.tocartesian(bz.toreduced([0.5, 0, 0])))
        for k in bz:
            assert np.allclose(k, np.zeros(3))

        w = 0.0
        for k, wk in bz.iter(True):
            assert np.allclose(k, np.zeros(3))
            w += wk
        assert w == pytest.approx(1.0)

        bz = BrillouinZone(setup.s1, [[0] * 3, [0.5] * 3], [0.5] * 2)
        assert len(bz) == 2
        assert len(bz.copy()) == 2

    def test_weight_automatic(self, setup):
        bz = BrillouinZone(1.0)
        assert bz.weight[0] == 1.0

        bz = BrillouinZone(setup.s1, np.random.rand(3, 3))
        assert bz.weight.sum() == pytest.approx(1)

        bz = BrillouinZone(setup.s1, np.random.rand(3, 3), 0.5)
        assert bz.weight.sum() == pytest.approx(1.5)

    def test_volume_self(self):
        bz = BrillouinZone(1.0)
        bz.parent.pbc = (False, False, False)
        v, dim = bz.volume(True)
        assert dim == 0
        assert v == pytest.approx(0)
        bz.parent.pbc = (True, False, False)
        assert bz.volume(True)[1] == 1
        bz.parent.pbc = (True, True, False)
        assert bz.volume(True)[1] == 2
        bz.parent.pbc = (True, True, True)
        v, dim = bz.volume(True)
        assert dim == 3
        assert v == pytest.approx((2 * np.pi) ** 3)

    def test_volume_direct(self):
        bz = BrillouinZone(1.0)
        assert bz.volume(True, [0, 1])[1] == 2
        assert bz.volume(True, [1])[1] == 1
        assert bz.volume(True, [2, 1])[1] == 2
        assert bz.volume(True, [2, 1, 0])[1] == 3
        assert bz.volume(True, [])[1] == 0

    def test_fail(self, setup):
        with pytest.raises(ValueError):
            BrillouinZone(setup.s1, [0] * 3, [0.5] * 2)

    def test_to_reduced(self, setup):
        bz = BrillouinZone(setup.s2)
        for k in [[0.1] * 3, [0.2] * 3]:
            cart = bz.tocartesian(k)
            rec = bz.toreduced(cart)
            assert np.allclose(rec, k)

    def test_class1(self, setup):
        class Test(LatticeChild):
            def __init__(self, lattice):
                self.set_lattice(lattice)

            def eigh(self, k, *args, **kwargs):
                return np.arange(3)

            def eig(self, k, *args, **kwargs):
                return np.arange(3) - 1

        bz = BrillouinZone(Test(setup.s1))
        bz_arr = bz.apply.array
        str(bz)
        assert np.allclose(bz_arr.eigh(), np.arange(3))
        assert np.allclose(bz_arr.eig(), np.arange(3) - 1)

    def test_class2(self, setup):
        class Test(LatticeChild):
            def __init__(self, lattice):
                self.set_lattice(lattice)

            def eigh(self, k, *args, **kwargs):
                return np.arange(3)

            def eig(self, k, *args, **kwargs):
                return np.arange(3) - 1

        bz = BrillouinZone(Test(setup.s1))
        # Try the list/yield method
        for val in bz.apply.list.eigh():
            assert np.allclose(val, np.arange(3))
        for val in bz.apply.iter.eigh():
            assert np.allclose(val, np.arange(3))
        for val in bz.apply.iter.eig():
            assert np.allclose(val, np.arange(3) - 1)
        for val in bz.apply.oplist.eigh():
            assert np.allclose(val, np.arange(3))
        # Average
        bz_average = bz.apply.average
        assert np.allclose(bz_average.eigh(), np.arange(3))
        assert np.allclose(bz_average.eigh(eta=True), np.arange(3))
        assert np.allclose(bz.apply.eigh.average(eta=True), np.arange(3))

    def test_parametrize_integer(self, setup):
        # parametrize for single integers
        def func(parent, N, i):
            return [i / N, 0, 0]

        bz = BrillouinZone.parametrize(setup.s1, func, 10)
        assert len(bz) == 10
        assert np.allclose(bz.k[-1], [9 / 10, 0, 0])

    def test_parametrize_list(self, setup):
        # parametrize for single integers
        def func(parent, N, i):
            return [i[0] / N[0], i[1] / N[1], 0]

        bz = BrillouinZone.parametrize(setup.s1, func, [10, 2])
        assert len(bz) == 20
        assert np.allclose(bz.k[-1], [9 / 10, 1 / 2, 0])
        assert np.allclose(bz.k[-2], [9 / 10, 0 / 2, 0])

    def test_default_weight(self):
        bz1 = BrillouinZone(geom.graphene(), [[0] * 3, [0.25] * 3], [1 / 2] * 2)
        bz2 = BrillouinZone(geom.graphene(), [[0] * 3, [0.25] * 3])
        assert np.allclose(bz1.k, bz2.k)
        assert np.allclose(bz1.weight, bz2.weight)

    def test_pickle(self, setup):
        import pickle as p

        bz1 = BrillouinZone(geom.graphene(), [[0] * 3, [0.25] * 3], [1 / 2] * 2)
        n = p.dumps(bz1)
        bz2 = p.loads(n)
        assert np.allclose(bz1.k, bz2.k)
        assert np.allclose(bz1.weight, bz2.weight)
        assert bz1.parent == bz2.parent

    @pytest.mark.parametrize("n", [[0, 0, 1], [0.5] * 3])
    def test_param_circle(self, n):
        bz = BrillouinZone.param_circle(1, 10, 0.1, n, [1 / 2] * 3)
        assert len(bz) == 10
        sc = Lattice(1)
        bz_loop = BrillouinZone.param_circle(sc, 10, 0.1, n, [1 / 2] * 3, True)
        assert len(bz_loop) == 10
        assert not np.allclose(bz.k, bz_loop.k)
        assert np.allclose(bz_loop.k[0, :], bz_loop.k[-1, :])

    def test_merge_simple(self):
        normal = [0] * 3
        origin = [1 / 2] * 3

        bzs = [
            BrillouinZone.param_circle(1, 10, 0.1, normal, origin),
            BrillouinZone.param_circle(1, 10, 0.2, normal, origin),
            BrillouinZone.param_circle(1, 10, 0.3, normal, origin),
        ]
        bz = BrillouinZone.merge(bzs)
        assert len(bz) == 30
        assert bz.weight.sum() == pytest.approx(
            bzs[0].weight.sum() + bzs[1].weight.sum() + bzs[2].weight.sum()
        )

    def test_merge_scales(self):
        normal = [0] * 3
        origin = [1 / 2] * 3

        bzs = [
            BrillouinZone.param_circle(1, 10, 0.1, normal, origin),
            BrillouinZone.param_circle(1, 10, 0.2, normal, origin),
            BrillouinZone.param_circle(1, 10, 0.3, normal, origin),
        ]
        bz = BrillouinZone.merge(bzs, [1, 2, 3])
        assert len(bz) == 30
        assert bz.weight.sum() == pytest.approx(
            bzs[0].weight.sum() + bzs[1].weight.sum() * 2 + bzs[2].weight.sum() * 3
        )

    def test_merge_scales_short(self):
        normal = [0] * 3
        origin = [1 / 2] * 3

        bzs = [
            BrillouinZone.param_circle(1, 10, 0.1, normal, [1 / 2] * 3),
            BrillouinZone.param_circle(1, 10, 0.2, normal, [1 / 2] * 3),
            BrillouinZone.param_circle(1, 10, 0.3, normal, [1 / 2] * 3),
        ]
        bz = BrillouinZone.merge(bzs, [1, 2])
        assert len(bz) == 30
        assert bz.weight.sum() == pytest.approx(
            bzs[0].weight.sum() + bzs[1].weight.sum() * 2 + bzs[2].weight.sum() * 2
        )

    def test_merge_scales_scalar(self):
        normal = [0] * 3
        origin = [1 / 2] * 3

        bzs = [
            BrillouinZone.param_circle(1, 10, 0.1, normal, [1 / 2] * 3),
            BrillouinZone.param_circle(1, 10, 0.3, normal, [1 / 2] * 3),
        ]
        bz = BrillouinZone.merge(bzs, 1)
        assert len(bz) == 20
        assert bz.weight.sum() == pytest.approx(
            bzs[0].weight.sum() + bzs[1].weight.sum()
        )


@pytest.mark.monkhorstpack
class TestMonkhorstPack:
    def test_class(self, setup):
        class Test(LatticeChild):
            def __init__(self, lattice):
                self.set_lattice(lattice)

            def eigh(self, k, *args, **kwargs):
                return np.arange(3)

            def eig(self, k, *args, **kwargs):
                return np.arange(3) - 1

        bz = MonkhorstPack(Test(setup.s1), [2] * 3)
        # Try the yield method
        bz_yield = bz.apply.iter
        for val in bz_yield.eigh():
            assert np.allclose(val, np.arange(3))
        for val in bz_yield.eig():
            assert np.allclose(val, np.arange(3) - 1)
        # Average
        assert np.allclose(bz.apply.average.eigh(), np.arange(3))

    def test_pickle(self, setup):
        import pickle as p

        bz1 = MonkhorstPack(geom.graphene(), [10, 11, 1], centered=False)
        n = p.dumps(bz1)
        bz2 = p.loads(n)
        assert np.allclose(bz1.k, bz2.k)
        assert np.allclose(bz1.weight, bz2.weight)
        assert bz1.parent == bz2.parent
        assert bz1._centered == bz2._centered

    @pytest.mark.parametrize("N", [2, 3, 4, 5, 7])
    @pytest.mark.parametrize("centered", [True, False])
    def test_asgrid(self, setup, N, centered):
        class Test(LatticeChild):
            def __init__(self, lattice):
                self.set_lattice(lattice)

            def eigh(self, k, *args, **kwargs):
                return np.arange(3)

        bz = MonkhorstPack(Test(setup.s1), [2] * 3).apply.grid

        # Check the shape
        grid = bz.eigh(wrap=lambda eig: eig[0])
        assert np.allclose(grid.shape, [2] * 3)

        # Check the grids are different
        grid2 = bz.eigh(grid_unit="Bohr", wrap=lambda eig: eig[0])
        assert not np.allclose(grid.cell, grid2.cell)

        assert np.allclose(grid.grid, grid2.grid)
        for i in range(3):
            grid = bz.eigh(data_axis=i)
            shape = [2] * 3
            shape[i] = 3
            assert np.allclose(grid.shape, shape)

    def test_asgrid_fail(self, setup):
        class Test(LatticeChild):
            def __init__(self, lattice):
                self.set_lattice(lattice)

            def eigh(self, k, *args, **kwargs):
                return np.arange(3)

        bz = MonkhorstPack(Test(setup.s1), [2] * 3, displacement=[0.1] * 3).apply.grid
        with pytest.raises(SislError):
            bz.eigh(wrap=lambda eig: eig[0])

    def test_init_simple(self, setup):
        bz = MonkhorstPack(setup.s1, [2] * 3, trs=False)
        assert len(bz) == 8
        assert bz.weight[0] == 1.0 / 8

    def test_displaced(self, setup):
        bz1 = MonkhorstPack(setup.s1, [2] * 3, centered=False, trs=False)
        assert len(bz1) == 8
        bz2 = MonkhorstPack(setup.s1, [2] * 3, displacement=[0.5] * 3, trs=False)
        assert len(bz2) == 8
        assert np.allclose(bz1.k, bz2.k)

    def test_uneven(self, setup):
        bz1 = MonkhorstPack(setup.s1, [3] * 3, trs=False)
        bz2 = MonkhorstPack(setup.s1, [3] * 3, displacement=[0.5] * 3, trs=False)
        assert not np.allclose(bz1.k, bz2.k)

    def test_size_half(self, setup):
        bz1 = MonkhorstPack(setup.s1, [2] * 3, size=0.5, trs=False)
        assert len(bz1) == 8
        assert np.all(bz1.k <= 0.25)
        assert bz1.weight.sum() == pytest.approx(0.5**3)

    def test_as_dataarray(self):
        pytest.importorskip("xarray", reason="xarray not available")

        from sisl import Hamiltonian, geom

        g = geom.graphene()
        H = Hamiltonian(g)
        H.construct([[0.1, 1.44], [0, -2.7]])

        bz = MonkhorstPack(H, [2, 2, 2], trs=False)

        # Assert that as* all does the same
        asarray = bz.apply.array.eigh()
        bz_da = bz.apply.dataarray
        asdarray = bz_da.eigh()
        assert np.allclose(asarray, asdarray.values)
        assert isinstance(asdarray.bz, MonkhorstPack)
        assert isinstance(asdarray.parent, Hamiltonian)
        assert asdarray.dims == ("k", "v1")

        asdarray = bz_da.eigh(coords=["orb"])
        assert asdarray.dims == ("k", "orb")

    def test_trs(self, setup):
        size = [0.05, 0.5, 0.9]
        for x, y, z in product(np.arange(10) + 1, np.arange(20) + 1, np.arange(6) + 1):
            bz = MonkhorstPack(setup.s1, [x, y, z])
            assert bz.weight.sum() == pytest.approx(1.0)
            bz = MonkhorstPack(setup.s1, [x, y, z], size=size)
            assert bz.weight.sum() == pytest.approx(np.prod(size))

    def test_gamma_centered(self, setup):
        for x, y, z in product(np.arange(10) + 1, np.arange(20) + 1, np.arange(6) + 1):
            bz = MonkhorstPack(setup.s1, [x, y, z], trs=False)
            assert len(bz) == x * y * z
            assert ((bz.k == 0.0).sum(1).astype(np.int32) == 3).sum() == 1

    def test_gamma_non_centered(self, setup):
        for x, y, z in product(np.arange(10) + 1, np.arange(20) + 1, np.arange(6) + 1):
            bz = MonkhorstPack(setup.s1, [x, y, z], centered=False, trs=False)
            assert len(bz) == x * y * z
            # The gamma point will also be in the unit-cell for
            # non-centered
            has_gamma = x % 2 == 1
            has_gamma &= y % 2 == 1
            has_gamma &= z % 2 == 1
            if has_gamma:
                assert ((bz.k == 0.0).sum(1).astype(np.int32) == 3).sum() == 1
            else:
                assert ((bz.k == 0.0).sum(1).astype(np.int32) == 3).sum() == 0

    def test_gamma_centered_displ(self, setup):
        for x, y, z in product(np.arange(10) + 1, np.arange(20) + 1, np.arange(6) + 1):
            bz = MonkhorstPack(setup.s1, [x, y, z], displacement=[0.2, 0, 0], trs=False)
            k = bz.k.copy()
            k[:, 0] -= 0.2
            assert len(bz) == x * y * z
            if x % 2 == 1:
                assert ((k == 0.0).sum(1).astype(np.int32) == 3).sum() == 1
            else:
                assert ((k == 0.0).sum(1).astype(np.int32) == 3).sum() == 0

    def test_as_simple(self):
        from sisl import Hamiltonian, geom

        g = geom.graphene()
        H = Hamiltonian(g)
        H.construct([[0.1, 1.44], [0, -2.7]])

        bz = MonkhorstPack(H, [2, 2, 2], trs=False)
        assert len(bz) == 2**3

        # Assert that as* all does the same
        apply = bz.apply
        asarray = apply.array.eigh()
        aslist = np.array(apply.list.eigh())
        asyield = np.array([a for a in apply.iter.eigh()])
        asaverage = apply.average.eigh()
        assert np.allclose(asarray, aslist)
        assert np.allclose(asarray, asyield)
        # Average needs to be performed
        assert np.allclose((asarray / len(bz)).sum(0), asaverage)
        apply.none.eigh()

    def test_as_dataarray_zip(self):
        pytest.importorskip("xarray", reason="xarray not available")

        from sisl import Hamiltonian, geom

        g = geom.graphene()
        H = Hamiltonian(g)
        H.construct([[0.1, 1.44], [0, -2.7]])

        E = np.linspace(-2, 2, 20)
        bz = MonkhorstPack(H, [2, 2, 1], trs=False)

        def wrap(es):
            return es.eig, es.DOS(E), es.PDOS(E)

        with bz.apply.renew(zip=True) as unzip:
            eig, DOS, PDOS = unzip.ndarray.eigenstate(wrap=wrap)
            ds0 = unzip.dataarray.eigenstate(wrap=wrap, name=["eig", "DOS", "PDOS"])
            # explicitly create dimensions
            ds1 = unzip.dataarray.eigenstate(
                wrap=wrap,
                coords=[
                    {"orb": np.arange(len(H))},
                    {"E": E},
                    {"spin": [0], "orb": np.arange(len(H)), "E": E},
                ],
                dims=(["orb"], ["E"], ["spin", "orb", "E"]),
                name=["eig", "DOS", "PDOS"],
            )

        for var, data in zip(["eig", "DOS", "PDOS"], [eig, DOS, PDOS]):
            assert np.allclose(ds0.data_vars[var].values, data)
            assert np.allclose(ds1.data_vars[var].values, data)
        assert len(ds1.coords) < len(ds0.coords)

    def test_bz_parallel_pathos(self):
        pytest.importorskip("pathos", reason="pathos not available")

        from sisl import Hamiltonian, geom

        g = geom.graphene()
        H = Hamiltonian(g)
        H.construct([[0.1, 1.44], [0, -2.7]])

        bz = MonkhorstPack(H, [2, 2, 2], trs=False)

        # try and determine a sensible
        import os

        nprocs = 2

        try:
            nprocs = len(os.sched_getaffinity(0)) // 2
        except Exception:
            pass

        try:
            import psutil

            nprocs = len(psutil.Process().cpu_affinity()) // 2
        except Exception:
            pass

        if nprocs == 1:
            pytest.skip("not in a parallel environment")

        nprocs = max(2, nprocs)
        omp_num_threads = os.environ.get("OMP_NUM_THREADS")

        # Check that the ObjectDispatcher works
        apply = bz.apply

        papply = bz.apply.renew(pool=nprocs)
        assert str(apply) != str(papply)

        for method in ("iter", "average", "sum", "array", "list", "oplist"):
            # TODO One should be careful with zip
            # zip will stop when it hits the final element in the first
            # list.
            # So if a generator has some clean-up code one has to use zip_longest
            # regardless of method
            os.environ["OMP_NUM_THREADS"] = "1"
            V1 = papply[method].eigh()
            if omp_num_threads is None:
                del os.environ["OMP_NUM_THREADS"]
            else:
                os.environ["OMP_NUM_THREADS"] = omp_num_threads
            V2 = apply[method].eigh()

            for v1, v2 in zip(V1, V2):
                assert np.allclose(v1, v2)

        # Check that the MethodDispatcher works
        apply = bz.apply.eigh
        papply = bz.apply.renew(pool=True).eigh
        assert str(apply) != str(papply)

        for method in ("iter", "average", "sum", "array", "list", "oplist"):
            # TODO One should be careful with zip
            # zip will stop when it hits the final element in the first
            # list.
            # So if a generator has some clean-up code one has to use zip_longest
            # regardless of method
            for v1, v2 in zip(papply[method](), apply[method]()):
                assert np.allclose(v1, v2)

    def test_as_single(self):
        from sisl import Hamiltonian, geom

        g = geom.graphene()
        H = Hamiltonian(g)
        H.construct([[0.1, 1.44], [0, -2.7]])

        def wrap(eig):
            return eig[0]

        bz = MonkhorstPack(H, [2, 2, 2], trs=False)
        assert len(bz) == 2**3

        # Assert that as* all does the same
        asarray = bz.apply.array.eigh(wrap=wrap)
        aslist = np.array(bz.apply.list.eigh(wrap=wrap))
        asyield = np.array([a for a in bz.apply.iter.eigh(wrap=wrap)])
        assert np.allclose(asarray, aslist)
        assert np.allclose(asarray, asyield)

    def test_as_wrap(self):
        from sisl import Hamiltonian, geom

        g = geom.graphene()
        H = Hamiltonian(g)
        H.construct([[0.1, 1.44], [0, -2.7]])

        bz = MonkhorstPack(H, [2, 2, 2], trs=False)
        assert len(bz) == 2**3

        # Check with a wrap function
        def wrap(arg):
            return arg[::-1]

        # Assert that as* all does the same
        asarray = bz.apply.array.eigh(wrap=wrap)
        aslist = np.array(bz.apply.list.eigh(wrap=wrap))
        asyield = np.array([a for a in bz.apply.iter.eigh(wrap=wrap)])
        asaverage = bz.apply.average.eigh(wrap=wrap)
        assert np.allclose(asarray, aslist)
        assert np.allclose(asarray, asyield)
        assert np.allclose((asarray / len(bz)).sum(0), asaverage)

        # Now we should check whether the reverse is doing its magic!
        mylist = [wrap(H.eigh(k=k)) for k in bz]
        assert np.allclose(aslist, mylist)

    def test_as_wrap_default_oplist(self):
        from sisl import Hamiltonian, geom

        g = geom.graphene()
        H = Hamiltonian(g)
        H.construct([[0.1, 1.44], [0, -2.7]])

        bz = MonkhorstPack(H, [2, 2, 2], trs=False)
        assert len(bz) == 2**3

        # Check with a wrap function
        E = np.linspace(-2, 2, 20)

        def wrap_sum(es, weight):
            PDOS = es.PDOS(E)[0] * weight
            return PDOS.sum(0), PDOS

        DOS, PDOS = bz.apply.sum.eigenstate(wrap=wrap_sum)
        bz_arr = bz.apply.array
        assert np.allclose(bz_arr.eigenstate(wrap=lambda es: es.DOS(E)), DOS)
        assert np.allclose(bz_arr.eigenstate(wrap=lambda es: es.PDOS(E)[0]), PDOS)

    def test_wrap_unzip(self):
        from sisl import Hamiltonian, geom

        g = geom.graphene()
        H = Hamiltonian(g)
        H.construct([[0.1, 1.44], [0, -2.7]])

        bz = MonkhorstPack(H, [2, 2, 2], trs=False)

        # Check with a wrap function
        E = np.linspace(-2, 2, 20)

        def wrap(es):
            return es.eig, es.DOS(E)

        eig0, DOS0 = zip(*bz.apply.list.eigenstate(wrap=wrap))
        with bz.apply.renew(zip=True) as k_unzip:
            eig1, DOS1 = k_unzip.list.eigenstate(wrap=wrap)
            eig2, DOS2 = k_unzip.array.eigenstate(wrap=wrap)

        # eig0 and DOS0 are generators, and not list's
        # eig1 and DOS1 are generators, and not list's
        assert isinstance(eig2, np.ndarray)
        assert isinstance(DOS2, np.ndarray)

        assert np.allclose(eig0, eig1)
        assert np.allclose(DOS0, DOS1)
        assert np.allclose(eig0, eig2)
        assert np.allclose(DOS0, DOS2)

    # Check with a wrap function and the weight argument
    def test_wrap_kwargs(arg):
        from sisl import Hamiltonian, geom

        g = geom.graphene()
        H = Hamiltonian(g)
        H.construct([[0.1, 1.44], [0, -2.7]])

        bz = MonkhorstPack(H, [2, 2, 2], trs=False)
        assert len(bz) == 2**3

        def wrap_none(arg):
            return arg

        def wrap_kwargs(arg, parent, k, weight):
            return arg * weight

        E = np.linspace(-2, 2, 20)
        bz_array = bz.apply.array
        asarray1 = (
            bz_array.eigenstate(wrap=lambda es: es.DOS(E)) * bz.weight.reshape(-1, 1)
        ).sum(0)
        asarray2 = bz_array.eigenstate(
            wrap=lambda es, parent, k, weight: es.DOS(E)
        ).sum(0)
        bz_list = bz.apply.list
        aslist1 = (
            np.array(bz_list.eigenstate(wrap=lambda es: es.DOS(E)))
            * bz.weight.reshape(-1, 1)
        ).sum(0)
        aslist2 = np.array(
            bz_list.eigenstate(wrap=lambda es, parent, k, weight: es.DOS(E))
        ).sum(0)
        bz_yield = bz.apply.iter
        asyield1 = (
            np.array([a for a in bz_yield.eigenstate(wrap=lambda es: es.DOS(E))])
            * bz.weight.reshape(-1, 1)
        ).sum(0)
        asyield2 = np.array(
            [
                a
                for a in bz_yield.eigenstate(
                    wrap=lambda es, parent, k, weight: es.DOS(E)
                )
            ]
        ).sum(0)

        asaverage = bz.apply.average.eigenstate(wrap=lambda es: es.DOS(E))
        assum = bz.apply.sum.eigenstate(wrap=lambda es: es.DOS(E))

        assert np.allclose(asarray1, asaverage)
        assert np.allclose(asarray2, asaverage)
        assert np.allclose(aslist1, asaverage)
        assert np.allclose(aslist2, asaverage)
        assert np.allclose(asyield1, asaverage)
        assert np.allclose(asyield2, asaverage)
        assert np.allclose(assum, asaverage)

    def test_replace_gamma(self):
        g = geom.graphene()
        bz = MonkhorstPack(g, 2, trs=False)
        bz_gamma = MonkhorstPack(g, [2, 2, 2], size=[0.5] * 3, trs=False)
        assert len(bz) == 2**3
        bz.replace([0] * 3, bz_gamma)
        assert len(bz) == 2**3 + 2**3 - 1
        assert bz.weight.sum() == pytest.approx(1.0)
        assert np.allclose(bz.copy().k, bz.k)
        assert np.allclose(bz.copy().weight, bz.weight)

    def test_replace_gamma_trs(self):
        g = geom.graphene()
        bz = MonkhorstPack(g, [2, 2, 2], trs=False)
        N_bz = len(bz)
        bz_gamma = MonkhorstPack(g, [3, 3, 3], size=[0.5] * 3, trs=True)
        N_bz_gamma = len(bz_gamma)
        bz.replace([0] * 3, bz_gamma)
        assert len(bz) == N_bz + N_bz_gamma - 1
        assert bz.weight.sum() == pytest.approx(1.0)

    def test_replace_trs_neg(self):
        g = geom.graphene()
        bz_big = MonkhorstPack(g, [6, 6, 1], trs=True)
        N_bz_big = len(bz_big)
        bz_small = MonkhorstPack(
            g,
            [3, 3, 3],
            size=[1 / 6, 1 / 6, 1],
            displacement=[2 / 3, 1 / 3, 0],
            trs=True,
        )
        N_bz_small = len(bz_small)

        # it should be the same for both negative|positive displ
        bz_pos = bz_big.copy()
        bz_neg = bz_big.copy()

        bz_pos.replace(bz_small.displacement, bz_small)
        bz_neg.replace(-bz_small.displacement, bz_small)
        for bz in [bz_pos, bz_neg]:
            assert len(bz) == N_bz_big + N_bz_small - 1
            assert bz.weight.sum() == pytest.approx(1.0)

    def test_in_primitive(self):
        assert np.allclose(MonkhorstPack.in_primitive([[1.0] * 3, [-1.0] * 3]), 0)


@pytest.mark.bandstructure
class TestBandStructure:
    def test_pbz1(self, setup):
        bz = BandStructure(setup.s1, [[0] * 3, [0.5] * 3], 300)
        assert len(bz) == 300

        bz2 = BandStructure(setup.s1, [[0] * 2, [0.5] * 2], 300, ["A", "C"])
        assert len(bz) == 300

        bz3 = BandStructure(setup.s1, [[0] * 2, [0.5] * 2], [150])
        assert len(bz) == 300
        bz.lineartick()
        bz.lineark()
        bz.lineark(True)

    @pytest.mark.parametrize("n", range(3, 100, 10))
    def test_pbz2(self, setup, n):
        bz = BandStructure(setup.s1, [[0] * 3, [0.25] * 3, [0.5] * 3], n)
        assert len(bz) == n

    def test_pbs_divisions(self, setup):
        bz = BandStructure(setup.s1, [[0] * 3, [0.25] * 3, [0.5] * 3], [10, 10])
        assert len(bz) == 21

    def test_pbc_fill(self, setup):
        bz = BandStructure(setup.s3, [[0] * 2, [0.25] * 2, [0.5] * 2], [10, 10])
        assert np.allclose(bz.k[:, 1], 0)
        assert len(bz) == 21

    def test_pbs_missing_arguments(self, setup):
        with pytest.raises(ValueError):
            bz = BandStructure(setup.s1, divisions=[10, 10])

    def test_pbs_fail(self, setup):
        with pytest.raises(ValueError):
            BandStructure(setup.s1, [[0] * 3, [0.5] * 3, [0.25] * 3], 1)
        with pytest.raises(ValueError):
            BandStructure(setup.s1, [[0] * 3, [0.5] * 3, [0.25] * 3], [1, 1, 1, 1])
        with pytest.raises(ValueError):
            BandStructure(setup.s1, [[0] * 3, [0.5] * 3, [0.25] * 3], [1, 1, 1])

    def test_pickle(self, setup):
        import pickle as p

        bz1 = BandStructure(setup.s1, [[0] * 2, [0.5] * 2], 300, ["A", "C"])
        n = p.dumps(bz1)
        bz2 = p.loads(n)
        assert np.allclose(bz1.k, bz2.k)
        assert np.allclose(bz1.weight, bz2.weight)
        assert bz1.parent == bz2.parent
        assert np.allclose(bz1.points, bz2.points)
        assert np.allclose(bz1.divisions, bz2.divisions)
        assert bz1.names == bz2.names

    def test_jump(self):
        g = geom.graphene()
        bs = BandStructure(
            g,
            [[0] * 3, [0.5, 0, 0], None, [0] * 3, [0.0, 0.5, 0]],
            30,
            ["A", "B", "C", "D"],
        )
        assert len(bs) == 30

    def test_jump_skipping_none(self):
        g = geom.graphene()
        bs1 = BandStructure(
            g,
            [[0] * 3, [0.5, 0, 0], None, [0] * 3, [0.0, 0.5, 0]],
            30,
            ["A", "B", "C", "D"],
        )
        bs2 = BandStructure(
            g,
            [[0] * 3, [0.5, 0, 0], None, [0] * 3, [0.0, 0.5, 0], None],
            30,
            ["A", "B", "C", "D"],
        )
        assert np.allclose(bs1.k, bs2.k)

    def test_insert_jump(self):
        g = geom.graphene()
        nk = 10
        bs = BandStructure(
            g,
            [[0] * 3, [0.5, 0, 0], None, [0] * 3, None, [0.0, 0.5, 0]],
            nk,
            ["A", "B", "C", "D"],
        )
        d = np.empty([nk])
        d_jump = bs.insert_jump(d)
        assert d_jump.shape == (nk + 2,)

        d = np.empty([nk, 5])
        d_jump = bs.insert_jump(d)
        assert d_jump.shape == (nk + 2, 5)
        assert np.isnan(d_jump).sum() == 10

        d_jump = bs.insert_jump(d.T, value=np.inf)
        assert d_jump.shape == (5, nk + 2)
        assert np.isinf(d_jump).sum() == 10

    def test_insert_jump_fail(self):
        g = geom.graphene()
        nk = 10
        bs = BandStructure(
            g,
            [[0] * 3, [0.5, 0, 0], None, [0] * 3, [0.0, 0.5, 0]],
            nk,
            ["A", "B", "C", "D"],
        )
        d = np.empty([nk + 1])
        with pytest.raises(ValueError):
            bs.insert_jump(d)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl import Atom, Geometry, Grid, Lattice, get_sile
from sisl._core.grid import sgrid


@pytest.fixture
def setup():
    class t:
        def __init__(self):
            bond = 1.42
            sq3h = 3.0**0.5 * 0.5
            self.lattice = Lattice(
                np.array(
                    [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
                )
                * bond,
                nsc=[3, 3, 1],
            )
            C = Atom(Z=6, R=[bond * 1.01] * 2)
            self.g = Geometry(
                np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
                atoms=C,
                lattice=self.lattice,
            )
            self.grid = Grid(0.2, geometry=self.g)
            self.grid.grid[:, :, :] = np.random.rand(*self.grid.shape)

            self.mol = Geometry([[i, 0, 0] for i in range(10)], lattice=[50])

            self.grid_mol = Grid(0.2, geometry=self.mol)
            self.grid_mol.grid[:, :, :] = np.random.rand(*self.grid_mol.shape)

            def sg_g(**kwargs):
                kwargs["ret_grid"] = True
                if "grid" not in kwargs:
                    kwargs["grid"] = self.grid
                return sgrid(**kwargs)

            self.sg_g = sg_g

            def sg_mol(**kwargs):
                kwargs["ret_grid"] = True
                if "grid" not in kwargs:
                    kwargs["grid"] = self.grid_mol
                return sgrid(**kwargs)

            self.sg_mol = sg_mol

    return t()


@pytest.mark.sgrid
class TestsGrid:
    def test_help(self):
        with pytest.raises(SystemExit):
            sgrid(argv=["--help"])

    def test_version(self):
        sgrid(argv=["--version"])

    def test_cite(self):
        sgrid(argv=["--cite"])

    def test_average1(self, setup):
        g = setup.grid.copy()
        gavg = g.average(0)
        for avg in ["average x", "average 0"]:
            G = setup.sg_g(argv=("--" + avg).split())
            assert np.allclose(G.grid, gavg.grid)
        gavg = g.average(1)
        for avg in ["average y", "average 1"]:
            G = setup.sg_g(argv=("--" + avg).split())
            assert np.allclose(G.grid, gavg.grid)
        gavg = g.average(2)
        for avg in ["average z", "average 2"]:
            G = setup.sg_g(argv=("--" + avg).split())
            assert np.allclose(G.grid, gavg.grid)

    def test_average2(self, setup):
        g = setup.grid.copy()
        gavg = g.average(0).average(1)
        for avg in ["average x --average 1", "average 0 --average y"]:
            G = setup.sg_g(argv=("--" + avg).split())
            assert np.allclose(G.grid, gavg.grid)

    def test_sum1(self, setup):
        g = setup.grid.copy()
        gavg = g.sum(0)
        for avg in ["sum x", "sum 0"]:
            G = setup.sg_g(argv=("--" + avg).split())
            assert np.allclose(G.grid, gavg.grid)
        gavg = g.sum(1)
        for avg in ["sum y", "sum 1"]:
            G = setup.sg_g(argv=("--" + avg).split())
            assert np.allclose(G.grid, gavg.grid)
        gavg = g.sum(2)
        for avg in ["sum z", "sum 2"]:
            G = setup.sg_g(argv=("--" + avg).split())
            assert np.allclose(G.grid, gavg.grid)

    def test_sum2(self, setup):
        g = setup.grid.copy()
        gavg = g.sum(0).sum(1)
        for avg in ["sum x --sum 1", "sum 0 --sum y"]:
            G = setup.sg_g(argv=("--" + avg).split())
            assert np.allclose(G.grid, gavg.grid)

    def test_print1(self, setup):
        setup.sg_g(argv=["--info"])

    def test_interp(self, setup):
        g1 = setup.sg_g(argv="--interp 10 10 10".split())
        # last argument is default
        g2 = setup.sg_g(argv="--interp 10 10 10 1".split())
        assert np.allclose(g1.grid, g2.grid)
        g2 = setup.sg_g(argv="--interp 10 10 10 3".split())
        assert not np.allclose(g1.grid, g2.grid)
        g3 = setup.sg_g(argv="--interp 0.01 0.1 1. 3".split())

    def test_smooth(self, setup):
        g1 = setup.sg_g(argv=["--smooth"])
        g2 = setup.sg_g(argv="--smooth 0.7".split())
        assert np.allclose(g1.grid, g2.grid)
        g2 = setup.sg_g(argv="--smooth 1.".split())
        assert not np.allclose(g1.grid, g2.grid)

    def test_sub1(self, setup):
        g = setup.grid.copy()
        idx = g.index(1.0, 0)
        gs = g.sub_part(idx, 0, True)
        for sub in ["sub 1.: a", "sub 1.: 0"]:
            G = setup.sg_g(argv=("--" + sub).split())
            assert np.allclose(G.grid, gs.grid)

        idx = g.index(1.2, 1)
        gs = g.sub_part(idx, 1, True)
        for sub in ["sub 1.2: b", "sub 1.2: y"]:
            G = setup.sg_g(argv=("--" + sub).split())
            assert np.allclose(G.grid, gs.grid)

        idx = g.index(0.8, 2)
        gs = g.sub_part(idx, 2, False)
        for sub in ["sub :.8 2", "sub :0.8 z"]:
            G = setup.sg_g(argv=("--" + sub).split())
            assert np.allclose(G.grid, gs.grid)

    def test_remove1(self, setup):
        g = setup.grid.copy()
        idx = g.index(1.0, 0)
        gs = g.remove_part(idx, 0, True)
        for remove in ["remove 1.: a", "remove 1.: 0"]:
            G = setup.sg_g(argv=("--" + remove).split())
            assert np.allclose(G.grid, gs.grid)

        idx = g.index(1.2, 1)
        gs = g.remove_part(idx, 1, False)
        for remove in ["remove :1.2 b", "remove :1.2 y"]:
            G = setup.sg_g(argv=("--" + remove).split())
            assert np.allclose(G.grid, gs.grid)

        idx = g.index(0.8, 2)
        gs = g.remove_part(idx, 2, False)
        for remove in ["remove :.8 2", "remove :0.8 z"]:
            G = setup.sg_g(argv=("--" + remove).split())
            assert np.allclose(G.grid, gs.grid)

    def test_tile1(self, setup):
        g = setup.grid.copy()

        g2 = g.tile(2, 0)
        G = setup.sg_g(argv="--tile 2 a".split())
        assert np.allclose(G.grid, g2.grid)

        g2 = g.tile(2, 1)
        G = setup.sg_g(argv="--tile 2 y".split())
        assert np.allclose(G.grid, g2.grid)

        g2 = g.tile(2, 2)
        G = setup.sg_g(argv="--tile 2 c".split())
        assert np.allclose(G.grid, g2.grid)

    def test_write_data(self, setup, sisl_tmp):
        out = sisl_tmp("table.dat")
        G = setup.sg_g(argv=f"--sum 0 --average 1 --out {out}".split())
        dat = get_sile(out).read_data()
        assert np.allclose(dat[1, :], G.grid.ravel())

    def test_write_grid(self, setup, sisl_tmp):
        out = sisl_tmp("table.cube")
        G = setup.sg_g(argv=f"--sum 0 --out {out}".split())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import math as m

import numpy as np
import pytest
from pytest import approx

import sisl
import sisl.linalg as lin
from sisl import BoundaryCondition, Lattice, LatticeChild, SuperCell
from sisl.geom import graphene

pytestmark = [pytest.mark.supercell, pytest.mark.sc, pytest.mark.lattice]


@pytest.fixture
def setup():
    class t:
        def __init__(self):
            alat = 1.42
            sq3h = 3.0**0.5 * 0.5
            self.lattice = Lattice(
                np.array(
                    [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
                )
                * alat,
                nsc=[3, 3, 1],
            )

    return t()


class TestLattice:
    def test_str(self, setup):
        str(setup.lattice)
        str(setup.lattice)
        assert setup.lattice != "Not a Lattice"

    def test_nsc1(self, setup):
        lattice = setup.lattice.copy()
        lattice.set_nsc([5, 5, 0])
        assert np.allclose([5, 5, 1], lattice.nsc)
        assert len(lattice.sc_off) == np.prod(lattice.nsc)

    def test_nsc2(self, setup):
        lattice = setup.lattice.copy()
        lattice.set_nsc([0, 1, 0])
        assert np.allclose([1, 1, 1], lattice.nsc)
        assert len(lattice.sc_off) == np.prod(lattice.nsc)
        lattice.set_nsc(a=3)
        assert np.allclose([3, 1, 1], lattice.nsc)
        assert len(lattice.sc_off) == np.prod(lattice.nsc)
        lattice.set_nsc(b=3)
        assert np.allclose([3, 3, 1], lattice.nsc)
        assert len(lattice.sc_off) == np.prod(lattice.nsc)
        lattice.set_nsc(c=5)
        assert np.allclose([3, 3, 5], lattice.nsc)
        assert len(lattice.sc_off) == np.prod(lattice.nsc)

    def test_nsc3(self, setup):
        assert setup.lattice.sc_index([0, 0, 0]) == 0
        for s in range(setup.lattice.n_s):
            assert setup.lattice.sc_index(setup.lattice.sc_off[s, :]) == s
        arng = np.arange(setup.lattice.n_s)
        np.random.seed(42)
        np.random.shuffle(arng)
        sc_off = setup.lattice.sc_off[arng, :]
        assert np.all(setup.lattice.sc_index(sc_off) == arng)

    def test_nsc_fail_even(self, setup):
        with pytest.raises(ValueError):
            setup.lattice.set_nsc(a=2)

    def test_nsc_fail_even_and_odd(self, setup):
        with pytest.raises(ValueError):
            setup.lattice.set_nsc([1, 2, 3])

    def test_area1(self, setup):
        setup.lattice.area(0, 1)

    def test_fill(self, setup):
        sc = setup.lattice.swapaxes(1, 2)
        i = sc._fill([1, 1])
        assert i.dtype == np.int32
        i = sc._fill([1.0, 1.0])
        assert i.dtype == np.float64
        for dt in [np.int32, np.float32, np.float64, np.complex64]:
            i = sc._fill([1.0, 1.0], dt)
            assert i.dtype == dt
            i = sc._fill(np.ones([2], dt))
            assert i.dtype == dt

    def test_add_vacuum_direct(self, setup):
        sc = setup.lattice.copy()
        for i in range(3):
            s = sc.add_vacuum(10, i)
            ax = setup.lattice.cell[i, :]
            ax += ax / np.sum(ax**2) ** 0.5 * 10
            assert np.allclose(ax, s.cell[i, :])

    def test_add_vacuum_orthogonal(self, setup):
        sc = setup.lattice.copy()
        s = sc.add_vacuum(10, 2, orthogonal_to_plane=True)
        assert np.allclose(s.cell[2], [0, 0, sc.cell[2, 2] + 10])

        # now check for the skewed ones
        s = sc.add_vacuum(sc.length[0], 0, orthogonal_to_plane=True)
        assert (s.length[0] / sc.length[0] - 1) * m.cos(m.radians(30)) == pytest.approx(
            1.0
        )

        # now check for the skewed ones
        s = sc.add_vacuum(sc.length[1], 1, orthogonal_to_plane=True)
        assert (s.length[1] / sc.length[1] - 1) * m.cos(m.radians(30)) == pytest.approx(
            1.0
        )

    def test_add1(self, setup):
        sc = setup.lattice.copy()
        for R in range(1, 10):
            s = sc + R
            assert np.allclose(s.cell, sc.cell + np.diag([R] * 3))

    def test_rotation1(self, setup):
        rot = setup.lattice.rotate(180, [0, 0, 1])
        rot.cell[2, 2] *= -1
        assert np.allclose(-rot.cell, setup.lattice.cell)

        rot = setup.lattice.rotate(m.pi, [0, 0, 1], rad=True)
        rot.cell[2, 2] *= -1
        assert np.allclose(-rot.cell, setup.lattice.cell)

        rot = rot.rotate(180, [0, 0, 1])
        rot.cell[2, 2] *= -1
        assert np.allclose(rot.cell, setup.lattice.cell)

    def test_rotation2(self, setup):
        rot = setup.lattice.rotate(180, setup.lattice.cell[2, :])
        rot.cell[2, 2] *= -1
        assert np.allclose(-rot.cell, setup.lattice.cell)

        rot = setup.lattice.rotate(m.pi, setup.lattice.cell[2, :], rad=True)
        rot.cell[2, 2] *= -1
        assert np.allclose(-rot.cell, setup.lattice.cell)

        rot = rot.rotate(180, setup.lattice.cell[2, :])
        rot.cell[2, 2] *= -1
        assert np.allclose(rot.cell, setup.lattice.cell)

    @pytest.mark.parametrize("a,b", [[0, 1], [0, 2], [1, 2]])
    def test_swapaxes_lattice_vector(self, setup, a, b):
        lattice = setup.lattice
        sab = lattice.swapaxes(a, b)
        assert np.allclose(lattice.origin, sab.origin)
        assert np.allclose(lattice.nsc[[b, a]], sab.nsc[[a, b]])
        assert np.allclose(lattice.cell[[b, a]], sab.cell[[a, b]])

        # and string input
        sab = setup.lattice.swapaxes("abc"[a], "abc"[b])
        assert np.allclose(sab.cell[[b, a]], setup.lattice.cell[[a, b]])
        assert np.allclose(sab.origin, setup.lattice.origin)
        assert np.allclose(sab.nsc[[b, a]], setup.lattice.nsc[[a, b]])

    @pytest.mark.parametrize("a,b", [[0, 1], [0, 2], [1, 2]])
    def test_swapaxes_xyz(self, setup, a, b):
        lattice = setup.lattice
        sab = lattice.swapaxes(1, 2, "xyz")
        assert np.allclose(lattice.nsc, sab.nsc)
        assert np.allclose(lattice.origin[[b, a]], sab.origin[[a, b]])
        assert np.allclose(lattice.origin[[b, a]], sab.origin[[a, b]])

        # and string input
        sab = setup.lattice.swapaxes("xyz"[a], "xyz"[b])
        assert np.allclose(lattice.nsc, sab.nsc)
        assert np.allclose(lattice.origin[[b, a]], sab.origin[[a, b]])
        assert np.allclose(lattice.origin[[b, a]], sab.origin[[a, b]])

    def test_swapaxes_complicated(self, setup):
        # swap a couple of lattice vectors and cartesian coordinates
        a = "azby"
        b = "bxcz"
        # this will result in
        # 0. abc, xyz
        # 1. bac, xyz
        # 2. bac, zyx
        # 3. bca, zyx
        # 4. bca, zxy
        sab = setup.lattice.swapaxes(a, b)
        idx_abc = [1, 2, 0]
        idx_xyz = [2, 0, 1]
        assert np.allclose(sab.cell, setup.lattice.cell[idx_abc][:, idx_xyz])
        assert np.allclose(sab.origin, setup.lattice.origin[idx_xyz])
        assert np.allclose(sab.nsc, setup.lattice.nsc[idx_abc])

    def test_offset1(self, setup):
        off = setup.lattice.offset()
        assert np.allclose(off, [0, 0, 0])
        off = setup.lattice.offset([1, 1, 1])
        cell = setup.lattice.cell[:, :]
        assert np.allclose(off, cell[0, :] + cell[1, :] + cell[2, :])

    def test_sc_index1(self, setup):
        sc_index = setup.lattice.sc_index([0, 0, 0])
        assert sc_index == 0
        sc_index = setup.lattice.sc_index([0, 0, None])
        assert len(sc_index) == setup.lattice.nsc[2]

    def test_sc_index2(self, setup):
        sc_index = setup.lattice.sc_index([[0, 0, 0], [1, 1, 0]])
        s = str(sc_index)
        assert len(sc_index) == 2

    def test_sc_index3(self, setup):
        with pytest.raises(Exception):
            setup.lattice.sc_index([100, 100, 100])

    def test_vertices(self, setup):
        verts = setup.lattice.vertices()

        assert verts.shape == (2, 2, 2, 3)
        assert np.allclose(verts[0, 0, 0], [0, 0, 0])
        assert np.allclose(verts[1, 0, 0], setup.lattice.cell[0])
        assert np.allclose(verts[1, 1, 1], setup.lattice.cell.sum(axis=0))

    def test_untile1(self, setup):
        cut = setup.lattice.untile(2, 0)
        assert np.allclose(cut.cell[0, :] * 2, setup.lattice.cell[0, :])
        assert np.allclose(cut.cell[1, :], setup.lattice.cell[1, :])

    def test_creation1(self, setup):
        # full cell
        tmp1 = Lattice([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        # diagonal cell
        tmp2 = Lattice([1, 1, 1])
        # cell parameters
        tmp3 = Lattice([1, 1, 1, 90, 90, 90])
        tmp4 = Lattice([1])
        assert np.allclose(tmp1.cell, tmp2.cell)
        assert np.allclose(tmp1.cell, tmp3.cell)
        assert np.allclose(tmp1.cell, tmp4.cell)

    def test_creation_latticechild_dispatch(self, setup):
        # full cell
        class P(LatticeChild):
            def copy(self):
                a = P()
                a.set_lattice(setup.lattice)
                return a

        tmp1 = P()
        tmp1.set_lattice([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

        # test dispatch method
        tmp2 = Lattice.new(tmp1)
        assert tmp1 == tmp2

    def test_creation_latticechild(self, setup):
        # full cell
        class P(LatticeChild):
            def copy(self):
                a = P()
                a.set_lattice(setup.lattice)
                return a

        tmp1 = P()
        tmp1.set_lattice([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

        # diagonal cell
        tmp2 = P()
        tmp2.set_lattice([1, 1, 1])
        # cell parameters
        tmp3 = P()
        tmp3.set_lattice([1, 1, 1, 90, 90, 90])
        tmp4 = P()
        tmp4.set_lattice([1])
        tmp5 = P()
        tmp5.set_lattice([[1, 1, 1], [90, 90, 90]])
        assert np.allclose(tmp1.cell, tmp2.cell)
        assert np.allclose(tmp1.cell, tmp3.cell)
        assert np.allclose(tmp1.cell, tmp4.cell)
        assert np.allclose(tmp1.cell, tmp5.cell)
        assert len(tmp1.lattice._fill([0, 0, 0])) == 3
        assert len(tmp1.lattice._fill_sc([0, 0, 0])) == 3
        assert tmp1.lattice.is_orthogonal()
        for i in range(3):
            t2 = tmp2.lattice.add_vacuum(10, i)
            assert tmp1.cell[i, i] + 10 == t2.cell[i, i]

    def test_creation3(self, setup):
        with pytest.raises(ValueError):
            setup.lattice.tocell([3, 6])

    def test_creation4(self, setup):
        with pytest.raises(ValueError):
            setup.lattice.tocell([3, 4, 5, 6])

    def test_creation5(self, setup):
        with pytest.raises(ValueError):
            setup.lattice.tocell([3, 4, 5, 6, 7, 6, 7])

    def test_creation_rotate(self, setup):
        # cell parameters
        param = np.array([1, 2, 3, 45, 60, 80], np.float64)
        parama = param.copy()
        parama[3:] *= np.pi / 180

        lattice = Lattice(param)

        param.shape = (2, 3)
        parama.shape = (2, 3)

        assert np.allclose(param, np.array(lattice.parameters()))
        assert np.allclose(parama, np.array(lattice.parameters(True)))
        for ang in range(0, 91, 5):
            s = (
                lattice.rotate(ang, lattice.cell[0, :])
                .rotate(ang, lattice.cell[1, :])
                .rotate(ang, lattice.cell[2, :])
            )
            assert np.allclose(param, np.array(s.parameters()))
            assert np.allclose(parama, np.array(s.parameters(True)))

    def test_rcell(self, setup):
        # LAPACK inverse algorithm implicitly does
        # a transpose.
        rcell = lin.inv(setup.lattice.cell) * 2.0 * np.pi
        assert np.allclose(rcell.T, setup.lattice.rcell)
        assert np.allclose(rcell.T / (2 * np.pi), setup.lattice.icell)

    def test_icell(self, setup):
        assert np.allclose(setup.lattice.rcell, setup.lattice.icell * 2 * np.pi)

    def test_center1(self, setup):
        assert np.allclose(
            setup.lattice.center(), np.sum(setup.lattice.cell, axis=0) / 2
        )
        for i in [0, 1, 2]:
            assert np.allclose(setup.lattice.center(i), setup.lattice.cell[i, :] / 2)

    def test_pickle(self, setup):
        import pickle as p

        s = p.dumps(setup.lattice)
        n = p.loads(s)
        assert setup.lattice == n
        s = Lattice([1, 1, 1])
        assert setup.lattice != s

    def test_orthogonal(self, setup):
        assert not setup.lattice.is_orthogonal()

    def test_fit1(self, setup):
        g = graphene()
        gbig = g.repeat(40, 0).repeat(40, 1)
        gbig.xyz[:, :] += (np.random.rand(len(gbig), 3) - 0.5) * 0.01
        lattice = g.lattice.fit(gbig)
        assert np.allclose(lattice.cell, gbig.cell)

    def test_fit2(self, setup):
        g = graphene(orthogonal=True)
        gbig = g.repeat(40, 0).repeat(40, 1)
        gbig.xyz[:, :] += (np.random.rand(len(gbig), 3) - 0.5) * 0.01
        lattice = g.lattice.fit(gbig)
        assert np.allclose(lattice.cell, gbig.cell)

    def test_fit3(self, setup):
        g = graphene(orthogonal=True)
        gbig = g.repeat(40, 0).repeat(40, 1)
        gbig.xyz[:, :] += (np.random.rand(len(gbig), 3) - 0.5) * 0.01
        lattice = g.lattice.fit(gbig, axes=0)
        assert np.allclose(lattice.cell[0, :], gbig.cell[0, :])
        assert np.allclose(lattice.cell[1:, :], g.cell[1:, :])

    def test_fit4(self, setup):
        g = graphene(orthogonal=True)
        gbig = g.repeat(40, 0).repeat(40, 1)
        gbig.xyz[:, :] += (np.random.rand(len(gbig), 3) - 0.5) * 0.01
        lattice = g.lattice.fit(gbig, axes=[0, 1])
        assert np.allclose(lattice.cell[0:2, :], gbig.cell[0:2, :])
        assert np.allclose(lattice.cell[2, :], g.cell[2, :])

    def test_parallel1(self, setup):
        g = graphene(orthogonal=True)
        gbig = g.repeat(40, 0).repeat(40, 1)
        assert g.lattice.parallel(gbig.lattice)
        assert gbig.lattice.parallel(g.lattice)
        g = g.rotate(90, g.cell[0, :], what="abc")
        assert not g.lattice.parallel(gbig.lattice)

    def test_tile_multiply_orthogonal(self):
        lattice = graphene(orthogonal=True).lattice
        assert np.allclose(
            lattice.tile(3, 0).tile(2, 1).tile(4, 2).cell, (lattice * (3, 2, 4)).cell
        )
        assert np.allclose(lattice.tile(3, 0).tile(2, 1).cell, (lattice * [3, 2]).cell)
        assert np.allclose(
            lattice.tile(3, 0).tile(3, 1).tile(3, 2).cell, (lattice * 3).cell
        )

    def test_tile_multiply_non_orthogonal(self):
        lattice = graphene(orthogonal=False).lattice
        assert np.allclose(
            lattice.tile(3, 0).tile(2, 1).tile(4, 2).cell, (lattice * (3, 2, 4)).cell
        )
        assert np.allclose(lattice.tile(3, 0).tile(2, 1).cell, (lattice * [3, 2]).cell)
        assert np.allclose(
            lattice.tile(3, 0).tile(3, 1).tile(3, 2).cell, (lattice * 3).cell
        )

    def test_angle1(self, setup):
        g = graphene(orthogonal=True)
        gbig = g.repeat(40, 0).repeat(40, 1)
        assert g.lattice.angle(0, 1) == 90

    def test_angle2(self, setup):
        lattice = Lattice([1, 1, 1])
        assert lattice.angle(0, 1) == 90
        assert lattice.angle(0, 2) == 90
        assert lattice.angle(1, 2) == 90
        lattice = Lattice([[1, 1, 0], [1, -1, 0], [0, 0, 2]])
        assert lattice.angle(0, 1) == 90
        assert lattice.angle(0, 2) == 90
        assert lattice.angle(1, 2) == 90
        lattice = Lattice([[3, 4, 0], [4, 3, 0], [0, 0, 2]])
        assert lattice.angle(0, 1, rad=True) == approx(0.28379, abs=1e-4)
        assert lattice.angle(0, 2) == 90
        assert lattice.angle(1, 2) == 90

    def test_cell2length(self):
        gr = graphene(orthogonal=True)
        lattice = (gr * (40, 40, 1)).rotate(24, gr.cell[2, :]).lattice
        assert np.allclose(
            lattice.length, (lattice.cell2length(lattice.length) ** 2).sum(1) ** 0.5
        )
        assert np.allclose(1, (lattice.cell2length(1) ** 2).sum(0))

    def test_set_lattice_off_wrong_size(self, setup):
        lattice = setup.lattice.copy()
        with pytest.raises(ValueError):
            lattice.sc_off = np.zeros([10000, 3])


def test_zero_length():
    with pytest.warns(sisl.SislWarning):
        Lattice([1, 0, 0])


def _dot(u, v):
    """Dot product u . v"""
    return u[0] * v[0] + u[1] * v[1] + u[2] * v[2]


def test_plane1():
    lattice = Lattice([1] * 3)
    # Check point [0.5, 0.5, 0.5]
    pp = np.array([0.5] * 3)

    n, p = lattice.plane(0, 1, True)
    assert -0.5 == approx(_dot(n, pp - p))
    n, p = lattice.plane(0, 2, True)
    assert -0.5 == approx(_dot(n, pp - p))
    n, p = lattice.plane(1, 2, True)
    assert -0.5 == approx(_dot(n, pp - p))
    n, p = lattice.plane(0, 1, False)
    assert -0.5 == approx(_dot(n, pp - p))
    n, p = lattice.plane(0, 2, False)
    assert -0.5 == approx(_dot(n, pp - p))
    n, p = lattice.plane(1, 2, False)
    assert -0.5 == approx(_dot(n, pp - p))


def test_plane2():
    lattice = Lattice([1] * 3)
    # Check point [-0.5, -0.5, -0.5]
    pp = np.array([-0.5] * 3)

    n, p = lattice.plane(0, 1, True)
    assert 0.5 == approx(_dot(n, pp - p))
    n, p = lattice.plane(0, 2, True)
    assert 0.5 == approx(_dot(n, pp - p))
    n, p = lattice.plane(1, 2, True)
    assert 0.5 == approx(_dot(n, pp - p))
    n, p = lattice.plane(0, 1, False)
    assert -1.5 == approx(_dot(n, pp - p))
    n, p = lattice.plane(0, 2, False)
    assert -1.5 == approx(_dot(n, pp - p))
    n, p = lattice.plane(1, 2, False)
    assert -1.5 == approx(_dot(n, pp - p))


def test_tocuboid_simple():
    lattice = Lattice([1, 1, 1, 90, 90, 90])
    with pytest.warns(sisl.SislDeprecation):
        c1 = lattice.toCuboid()
    assert np.allclose(lattice.cell, c1._v)
    c1 = lattice.to.Cuboid()
    assert np.allclose(lattice.cell, c1._v)

    with pytest.warns(sisl.SislDeprecation):
        c2 = lattice.toCuboid(orthogonal=True)
    assert np.allclose(c1._v, c2._v)
    c2 = lattice.to.Cuboid(orthogonal=True)
    assert np.allclose(c1._v, c2._v)


def test_tocuboid_complex():
    lattice = Lattice([1, 1, 1, 60, 60, 60])
    s = str(lattice.cell)
    with pytest.warns(sisl.SislDeprecation):
        c1 = lattice.toCuboid()
    assert np.allclose(lattice.cell, c1._v)
    c1 = lattice.to.Cuboid()
    assert np.allclose(lattice.cell, c1._v)

    c1 = lattice.to.Cuboid(orthogonal=True)
    assert not np.allclose(np.diagonal(lattice.cell), np.diagonal(c1._v))
    with pytest.warns(sisl.SislDeprecation):
        c2 = lattice.toCuboid(orthogonal=True)
    assert np.allclose(np.diagonal(c1._v), np.diagonal(c2._v))


def test_fortran_contiguous():
    # for #748
    abc = np.zeros([3, 3], order="F")
    latt = Lattice(abc)
    assert latt.cell.flags.c_contiguous


def test_lattice_indices():
    lattice = Lattice([1] * 3, nsc=[3, 5, 7])
    for i in range(lattice.n_s):
        assert i == lattice.sc_index(lattice.sc_off[i])


def test_supercell_warn():
    with pytest.warns(sisl.SislDeprecation):
        lattice = SuperCell([1] * 3, nsc=[3, 5, 7])


#####
# boundary-condition tests
#####


@pytest.mark.parametrize("bc", list(BoundaryCondition))
def test_lattice_bc_init(bc):
    Lattice(1, boundary_condition=bc)
    Lattice(1, boundary_condition=[bc, bc, bc])
    Lattice(1, boundary_condition=[[bc, bc], [bc, bc], [bc, bc]])


def test_lattice_bc_set():
    lat = Lattice(1, boundary_condition=Lattice.BC.PERIODIC)
    assert lat.pbc.all()
    lat.boundary_condition = Lattice.BC.UNKNOWN
    assert not lat.pbc.any()
    assert (lat.boundary_condition == Lattice.BC.UNKNOWN).all()

    lat.boundary_condition = ["per", "unkn", [3, 4]]

    for n in "abc":
        lat.set_boundary_condition(**{n: "per"})
        lat.set_boundary_condition(**{n: [3, Lattice.BC.UNKNOWN]})
        lat.set_boundary_condition(**{n: [True, Lattice.BC.PERIODIC]})

    bc = ["per", ["Dirichlet", Lattice.BC.NEUMANN], ["un", "neu"]]
    lat.set_boundary_condition(bc)
    assert np.all(
        lat.boundary_condition[1] == [Lattice.BC.DIRICHLET, Lattice.BC.NEUMANN]
    )
    assert np.all(lat.boundary_condition[2] == [Lattice.BC.UNKNOWN, Lattice.BC.NEUMANN])
    lat.set_boundary_condition(["per", None, ["dirichlet", "unkno"]])
    assert np.all(
        lat.boundary_condition[1] == [Lattice.BC.DIRICHLET, Lattice.BC.NEUMANN]
    )
    assert np.all(
        lat.boundary_condition[2] == [Lattice.BC.DIRICHLET, Lattice.BC.UNKNOWN]
    )
    lat.set_boundary_condition("ppu")
    assert np.all(lat.pbc == [True, True, False])


def test_lattice_bc_fail():
    lat = Lattice(1)
    with pytest.raises(ValueError):
        lat.boundary_condition = False
    with pytest.raises(ValueError):
        lat.set_boundary_condition(b=False)
    with pytest.raises(KeyError):
        lat.set_boundary_condition(b="eusoatuhesoau")


def test_new_inherit_class():
    l = Lattice(1, boundary_condition=Lattice.BC.NEUMANN)

    class MyClass(Lattice):
        pass

    l2 = MyClass.new(l)
    assert l2.__class__ == MyClass
    assert l2 == l
    l2 = Lattice.new(l)
    assert l2.__class__ == Lattice
    assert l2 == l


def test_lattice_info():
    lat = Lattice(1, nsc=[3, 3, 3])
    with pytest.warns(sisl.SislWarning) as record:
        lat.set_boundary_condition(b=Lattice.BC.DIRICHLET)
        lat.set_boundary_condition(c=Lattice.BC.PERIODIC)
        assert len(record) == 1


def test_lattice_pbc_setter():
    lat = Lattice(1, nsc=[3, 3, 3])
    assert np.all(lat.pbc)
    pbc = [True, False, True]
    lat.pbc = pbc
    assert np.all(lat.pbc == pbc)
    # check that None won't change anything
    pbc = [False, None, None]
    lat.pbc = pbc
    assert np.all(lat.pbc == [False, False, True])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import pytest

approx = pytest.approx

import numpy as np

from sisl.unit import unit_convert, unit_default, unit_group, units

pytestmark = pytest.mark.unit


def test_group():
    assert unit_group("kg") == "mass"
    assert unit_group("eV") == "energy"
    assert unit_group("N") == "force"


def test_unittable():
    groups = units._table.groups()
    assert "mass" in groups
    assert "energy" in groups


def test_unit_convert():
    assert approx(unit_convert("kg", "g")) == 1.0e3
    assert approx(unit_convert("eV", "J")) == 1.60217733e-19
    assert approx(unit_convert("J", "eV")) == 1 / 1.60217733e-19
    assert approx(unit_convert("J", "eV", opts={"^": 2})) == (1 / 1.60217733e-19) ** 2
    assert approx(unit_convert("J", "eV", opts={"/": 2})) == (1 / 1.60217733e-19) / 2
    assert approx(unit_convert("J", "eV", opts={"*": 2})) == (1 / 1.60217733e-19) * 2
    assert approx(unit_convert("Ha", "Ry")) == 2
    assert approx(unit_convert("Ha", "eV")) == 27.211386245988
    assert approx(unit_convert("eV", "Hz")) == 2.417989242e14
    assert approx(unit_convert("eV", "MHz")) == 2.417989242e8
    assert approx(unit_convert("K", "eV")) == 8.617333262e-5
    assert approx(unit_convert("MHz", "Hz")) == 1e6
    assert approx(unit_convert("GHz", "Hz")) == 1e9
    assert approx(unit_convert("THz", "Hz")) == 1e12
    assert approx(unit_convert("invcm", "eV")) == 1.239841984e-4


def test_class_unit():
    assert np.allclose(units.convert("J", "J", "J"), 1)

    assert approx(units.convert("kg", "g")) == 1.0e3
    assert approx(units.convert("eV", "J")) == 1.60217733e-19
    assert approx(units.convert("J", "eV")) == 1 / 1.60217733e-19
    assert approx(units.convert("J^2", "eV**2")) == (1 / 1.60217733e-19) ** 2
    assert approx(units.convert("J/2", "eV/2")) == (1 / 1.60217733e-19)
    assert approx(units.convert("J", "eV/2")) == (1 / 1.60217733e-19) * 2
    assert approx(units.convert("J2", "2eV")) == (1 / 1.60217733e-19)
    assert approx(units.convert("J2", "eV")) == (1 / 1.60217733e-19) * 2
    assert approx(units.convert("J/m", "eV/Ang")) == unit_convert(
        "J", "eV"
    ) / unit_convert("m", "Ang")
    units("J**eV", "eV**eV")
    units("J/m", "eV/m")


@pytest.mark.xfail(reason="extracting math before checking units in tables")
def test_class_unit_math_in_table():
    assert np.allclose(units.convert("cm^-1", "eV"), 1.239841984e-4)


def test_default():
    assert unit_default("mass") == "amu"
    assert unit_default("length") == "Ang"
    assert unit_default("time") == "fs"
    assert unit_default("energy") == "eV"
    assert unit_default("force") == "eV/Ang"
    assert unit_default("velocity") == "Ang/fs"
    assert unit_default("pressure") == "eV/Ang^3"
    assert unit_default("efield") == "V/Ang"


def test_group_f1():
    with pytest.raises(ValueError):
        unit_group("not-existing")


def test_default_f1():
    with pytest.raises(ValueError):
        unit_default("not-existing")


def test_unit_convert_f1():
    with pytest.raises(ValueError):
        unit_convert("eV", "megaerg")


def test_unit_convert_f2():
    with pytest.raises(ValueError):
        unit_convert("eV", "kg")


def test_unit_convert_single():
    assert units("Ang") == 1e-10



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

import sisl as si
from sisl.io.ham import *

pytestmark = [pytest.mark.io, pytest.mark.generic]


def test_ham_geometry(sisl_tmp, sisl_system):
    f = sisl_tmp("gr.ham")
    sisl_system.g.write(hamiltonianSile(f, "w"))
    g = hamiltonianSile(f).read_geometry()

    # Assert they are the same
    assert np.allclose(g.cell, sisl_system.g.cell)
    assert np.allclose(g.xyz, sisl_system.g.xyz)
    assert g.atoms.equal(sisl_system.g.atoms, R=False)


def test_ham_nn(sisl_tmp, sisl_system):
    f = sisl_tmp("gr.ham")
    sisl_system.ham.write(hamiltonianSile(f, "w"))
    ham = hamiltonianSile(f).read_hamiltonian()
    assert ham.spsame(sisl_system.ham)


@pytest.mark.parametrize("hermitian", [True, False])
def test_ham_3nn(sisl_tmp, hermitian):
    f = sisl_tmp("gr.ham")

    g = si.geom.graphene()
    H = si.Hamiltonian(g)

    # build a 3rd nearest neighbor model
    H.construct([[0.1, 1.6, 2.6, 3.1], [0, -2.7, -0.20, -0.18]])

    H.write(f, hermitian=hermitian)

    H1 = hamiltonianSile(f).read_hamiltonian()
    assert np.allclose((H1 - H)._csr._D, 0)

    if not hermitian:
        H1 = hamiltonianSile(f).read_hamiltonian(hermitian=hermitian)
        assert np.allclose((H1 - H)._csr._D, 0)


@pytest.mark.parametrize("hermitian", [True, False])
def test_ham_3nn_no(sisl_tmp, hermitian):
    f = sisl_tmp("gr.ham")

    g = si.geom.graphene()
    g.cell[2, 2] = 1.00
    H = si.Hamiltonian(g, orthogonal=False)
    H.set_nsc([5, 5, 5])

    # build a 3rd nearest neighbor, nonorthogonal model
    H.construct(
        [[0.1, 1.6, 2.6, 3.1], [(1, 1), (-2.7, 0.073), (-0.09, 0.045), (-0.33, 0.026)]]
    )

    H.write(f, hermitian=hermitian)

    H1 = hamiltonianSile(f).read_hamiltonian()
    assert np.allclose((H1 - H)._csr._D, 0)

    if not hermitian:
        H1 = hamiltonianSile(f).read_hamiltonian(hermitian=hermitian)
        assert np.allclose((H1 - H)._csr._D, 0)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

import sisl
from sisl import (
    Atom,
    DensityMatrix,
    DynamicalMatrix,
    EnergyDensityMatrix,
    Geometry,
    Hamiltonian,
)
from sisl.io.siesta import *

pytestmark = [pytest.mark.io, pytest.mark.siesta]

netCDF4 = pytest.importorskip("netCDF4")


def test_nc1(sisl_tmp, sisl_system):
    f = sisl_tmp("gr.nc")
    tb = Hamiltonian(sisl_system.gtb)
    tb.construct([sisl_system.R, sisl_system.t])
    with ncSileSiesta(f, "w") as s:
        tb.write(s)

    with ncSileSiesta(f) as f:
        ntb = f.read_hamiltonian()

    # Assert they are the same
    assert np.allclose(tb.cell, ntb.cell)
    assert np.allclose(tb.xyz, ntb.xyz)
    tb.finalize()
    assert np.allclose(tb._csr._D[:, 0], ntb._csr._D[:, 0])
    assert sisl_system.g.atoms.equal(ntb.atoms, R=False)


def test_nc2(sisl_tmp, sisl_system):
    f = sisl_tmp("grS.nc")
    tb = Hamiltonian(sisl_system.gtb, orthogonal=False)
    tb.construct([sisl_system.R, sisl_system.tS])
    with ncSileSiesta(f, "w") as s:
        tb.write(s)

    with ncSileSiesta(f) as f:
        ntb = f.read_hamiltonian()

    # Assert they are the same
    assert np.allclose(tb.cell, ntb.cell)
    assert np.allclose(tb.xyz, ntb.xyz)
    tb.finalize()
    assert np.allclose(tb._csr._D[:, 0], ntb._csr._D[:, 0])
    assert sisl_system.g.atoms.equal(ntb.atoms, R=False)


def test_nc_multiple_fail(sisl_tmp, sisl_system):
    # writing two different sparse matrices to the same
    # file will fail
    f = sisl_tmp("gr.nc")
    H = Hamiltonian(sisl_system.gtb)
    DM = DensityMatrix(sisl_system.gtb)

    with ncSileSiesta(f, "w") as sile:
        H.construct([sisl_system.R, sisl_system.t])
        H.write(sile)

        DM[0, 0] = 1.0
        with pytest.raises(ValueError):
            with pytest.warns(sisl.SislWarning, match=r"changes the sparsity pattern"):
                DM.write(sile)


@pytest.mark.parametrize(
    ("sort"),
    [True, False],
)
def test_nc_multiple_checks(sisl_tmp, sisl_system, sort):
    f = sisl_tmp("gr.nc")
    H = Hamiltonian(sisl_system.gtb)
    DM = DensityMatrix(sisl_system.gtb)

    with ncSileSiesta(f, "w") as sile:
        H.construct([sisl_system.R, sisl_system.t])
        H.write(sile, sort=sort)

        # fix seed
        np.random.seed(42)
        shuffle = np.random.shuffle
        for io in range(len(H)):
            edges = H.edges(io)  # get all edges
            shuffle(edges)
            DM[io, edges] = 2.0

        DM.write(sile, sort=sort)


def test_nc_overlap(sisl_tmp, sisl_system):
    f = sisl_tmp("gr.nc")
    tb = Hamiltonian(sisl_system.gtb)
    tb.construct([sisl_system.R, sisl_system.t])
    tb.write(ncSileSiesta(f, "w"))

    with ncSileSiesta(f) as sile:
        S = sile.read_overlap()

    # Ensure no empty data-points
    assert np.allclose(S._csr._D.sum(), tb.no)

    # Write test
    f = sisl_tmp("s.nc")
    with ncSileSiesta(f, "w") as sile:
        S.write(sile)
    with ncSileSiesta(f) as sile:
        S2 = sile.read_overlap()
    assert S._csr.spsame(S2._csr)
    assert np.allclose(S._csr._D, S2._csr._D)


def test_nc_dynamical_matrix(sisl_tmp, sisl_system):
    f = sisl_tmp("grdyn.nc")
    dm = DynamicalMatrix(sisl_system.gtb)
    for _, ix in dm.iter_orbitals():
        dm[ix, ix] = ix / 2.0

    with ncSileSiesta(f, "w") as sile:
        dm.write(sile)

    with ncSileSiesta(f) as sile:
        ndm = sile.read_dynamical_matrix()

    # Assert they are the same
    assert np.allclose(dm.cell, ndm.cell)
    assert np.allclose(dm.xyz, ndm.xyz)
    dm.finalize()
    assert np.allclose(dm._csr._D[:, 0], ndm._csr._D[:, 0])
    assert sisl_system.g.atoms.equal(ndm.atoms, R=False)


def test_nc_density_matrix(sisl_tmp, sisl_system):
    f = sisl_tmp("grDM.nc")
    dm = DensityMatrix(sisl_system.gtb)
    for _, ix in dm.iter_orbitals():
        dm[ix, ix] = ix / 2.0

    with ncSileSiesta(f, "w") as sile:
        dm.write(sile)

    with ncSileSiesta(f) as sile:
        ndm = sile.read_density_matrix()

    # Assert they are the same
    assert np.allclose(dm.cell, ndm.cell)
    assert np.allclose(dm.xyz, ndm.xyz)
    dm.finalize()
    assert np.allclose(dm._csr._D[:, 0], ndm._csr._D[:, 0])
    assert sisl_system.g.atoms.equal(ndm.atoms, R=False)


@pytest.mark.filterwarnings("ignore", message="*is NOT Hermitian for on-site")
def test_nc_H_spin_orbit_nc2tshs2nc(sisl_tmp):
    H1 = Hamiltonian(sisl.geom.graphene(), spin=sisl.Spin("SO"))
    H1.construct(
        (
            [0.1, 1.44],
            [
                [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8],
                [0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9],
            ],
        )
    )

    f1 = sisl_tmp("H1.nc")
    f2 = sisl_tmp("H2.TSHS")
    H1.write(f1)
    H1.finalize()
    with sisl.get_sile(f1) as sile:
        H2 = sile.read_hamiltonian()
    H2.write(f2)
    with sisl.get_sile(f2) as sile:
        H3 = sile.read_hamiltonian()
    assert H1._csr.spsame(H2._csr)
    assert np.allclose(H1._csr._D, H2._csr._D)
    assert H1._csr.spsame(H3._csr)
    assert np.allclose(H1._csr._D, H3._csr._D)


@pytest.mark.filterwarnings("ignore", message="*is NOT Hermitian for on-site")
def test_nc_DM_spin_orbit_nc2dm2nc(sisl_tmp):
    DM1 = DensityMatrix(sisl.geom.graphene(), orthogonal=False, spin=sisl.Spin("SO"))
    DM1.construct(
        (
            [0.1, 1.44],
            [
                [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.0],
                [0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.0],
            ],
        )
    )

    f1 = sisl_tmp("DM1.nc")
    f2 = sisl_tmp("DM2.DM")
    DM1.finalize()
    DM1.write(f1)
    with sisl.get_sile(f1) as sile:
        DM2 = sile.read_density_matrix()
    DM2.write(f2)
    with sisl.get_sile(f2) as sile:
        DM3 = sile.read_density_matrix()
    assert DM1._csr.spsame(DM2._csr)
    assert np.allclose(DM1._csr._D, DM2._csr._D)
    assert DM1._csr.spsame(DM3._csr)
    assert np.allclose(DM1._csr._D, DM3._csr._D)


def test_nc_ghost(sisl_tmp):
    f = sisl_tmp("ghost.nc")
    a1 = Atom(1)
    am1 = Atom(-1)
    g = Geometry([[0.0, 0.0, i] for i in range(2)], [a1, am1], 2.0)
    g.write(ncSileSiesta(f, "w"))

    with ncSileSiesta(f) as sile:
        g2 = sile.read_geometry()
    assert np.allclose(g.cell, g2.cell)
    assert np.allclose(g.xyz, g2.xyz)
    assert np.allclose(g.atoms.Z, g2.atoms.Z)
    assert g.atoms[0].__class__ is g2.atoms[0].__class__
    assert g.atoms[1].__class__ is g2.atoms[1].__class__
    assert g.atoms[0].__class__ is not g2.atoms[1].__class__



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import numpy as np
import pytest

import sisl
from sisl import Grid
from sisl._sparse_grid_ops import transpose_raveled_index


@pytest.fixture
def geometry():

    r = np.linspace(0, 3.5, 50)
    f = np.exp(-r)

    orb = sisl.AtomicOrbital("2pzZ", (r, f))
    geom = sisl.geom.graphene(orthogonal=True, atoms=sisl.Atom(6, orb))
    geom.set_nsc([3, 5, 1])

    return geom


@pytest.fixture
def grid_shape():
    return (8, 10, 12)


@pytest.fixture
def psi_values(geometry, grid_shape):
    return geometry._orbital_values(grid_shape)


@pytest.fixture
def H(geometry):
    H = sisl.Hamiltonian(geometry)
    H.construct(
        [(0.1, 1.44), (0, -2.7)],
    )
    return H


def test_transpose_raveled():
    """Tests the cython implemented function that transposes
    raveled indices."""

    # Define the reference function, which calls numpy
    def transpose_raveled_numpy(index, grid_shape, new_order):
        grid_shape = np.array(grid_shape)

        unraveled = np.unravel_index(index, grid_shape)
        unraveled = np.array(unraveled)
        new_grid_shape = grid_shape[new_order]
        new_unraveled = unraveled[new_order]

        return np.ravel_multi_index(new_unraveled, new_grid_shape)

    grid_shape = np.array([3, 4, 5], dtype=np.int32)
    grid_size = np.prod(grid_shape)
    # Move the first axis to the last dimension.
    new_order = np.array([1, 2, 0], dtype=np.int32)

    for i in range(grid_size):
        transposed = transpose_raveled_index(i, grid_shape, new_order)
        transposed_np = transpose_raveled_numpy(i, grid_shape, new_order)

        assert transposed == transposed_np

    # Move the last axis to the first dimension.
    new_order = np.array([2, 0, 1], dtype=np.int32)

    for i in range(grid_size):
        transposed = transpose_raveled_index(i, grid_shape, new_order)
        transposed_np = transpose_raveled_numpy(i, grid_shape, new_order)

        assert transposed == transposed_np


@pytest.mark.parametrize("k", [(0, 0, 0), (0.5, 0, 0)])
def test_wavefunction_correct(H, grid_shape, psi_values, k):
    """Checks that the wavefunction computed with the precalculated
    psi values is the same as the wavefunction computed directly."""

    eig = H.eigenstate(k=k)[0]

    wf_grid = sisl.Grid(grid_shape, geometry=H.geometry, dtype=np.complex128)

    eig.wavefunction(wf_grid)

    from_psi_grid = psi_values.reduce_orbitals(eig.state.T, k=k)

    assert np.allclose(from_psi_grid.grid, wf_grid.grid)


@pytest.mark.parametrize(
    ["k", "ncoeffs"],
    [[(0, 0, 0), 1], [(0, 0, 0), 2], [(0.25, 0, 0), 1], [(0.25, 0, 0), 2]],
)
def test_onthefly_reduction(geometry, psi_values, k, ncoeffs):
    """Checks that the on the fly reduction produces the same
    results as computing the whole grid and then reducing."""

    coeffs = np.random.random((geometry.no, ncoeffs))

    not_reduced = psi_values.reduce_orbitals(coeffs, k=k)

    # Reducing the last axis
    reduced = psi_values.reduce_orbitals(coeffs, k=k, reduce_grid=(2,))

    post_reduced = not_reduced.sum(2)
    assert reduced.size == post_reduced.size
    if ncoeffs == 1:
        assert np.allclose(reduced.grid, post_reduced.grid)
    else:
        assert np.allclose(reduced.reshape(post_reduced.shape), post_reduced)

    # Reducing the two last axes
    reduced = psi_values.reduce_orbitals(coeffs, k=k, reduce_grid=(2, 1))

    post_reduced = not_reduced.sum(2).sum(1)
    assert reduced.size == post_reduced.size
    if ncoeffs == 1:
        assert np.allclose(reduced.grid, post_reduced.grid)
    else:
        assert np.allclose(reduced.reshape(post_reduced.shape), post_reduced)

    # Now we are going to reduce the leading axes. This is more involved, so
    # it is more likely to fail.

    # Reducing the first axis
    reduced = psi_values.reduce_orbitals(coeffs, k=k, reduce_grid=(0,))

    post_reduced = not_reduced.sum(0)
    assert reduced.size == post_reduced.size
    if ncoeffs == 1:
        assert np.allclose(reduced.grid, post_reduced.grid)
    else:
        assert np.allclose(reduced.reshape(post_reduced.shape), post_reduced)

    # Reducing the first and second axis
    reduced = psi_values.reduce_orbitals(coeffs, k=k, reduce_grid=(0, 1))

    post_reduced = not_reduced.sum(1).sum(0)
    assert reduced.size == post_reduced.size
    if ncoeffs == 1:
        assert np.allclose(reduced.grid, post_reduced.grid)
    else:
        assert np.allclose(reduced.reshape(post_reduced.shape), post_reduced)


def test_orbital_products(geometry):
    """Very simple tests to see if the orbital products are computed correctly"""
    geometry = geometry.copy()
    geometry.set_nsc([1, 1, 1])
    geometry.lattice.set_boundary_condition("open")

    # Don't get periodic contributions for orbital values, then predicting
    # grid values will be much easier.
    psi_values = geometry._orbital_values((10, 10, 10))

    orb_csr = psi_values._csr.tocsr()
    orb_0 = orb_csr[:, 0].toarray().ravel()
    orb_1 = orb_csr[:, 1].toarray().ravel()

    # Compute the orbital products with one coefficient
    DM = sisl.DensityMatrix(geometry, dim=1, dtype=np.float64)
    DM[0, 0] = 1.0

    dens = psi_values.reduce_orbital_products(DM.tocsr(), geometry.lattice)
    assert isinstance(dens, Grid)
    assert dens.shape == psi_values.grid_shape
    assert np.any(dens.grid != 0)

    predicted = orb_0**2

    assert np.allclose(dens.grid.ravel(), predicted)

    # Compute the orbital products with two diagonal coefficients
    DM = sisl.DensityMatrix(geometry, dim=1, dtype=np.float64)
    DM[0, 0] = 1.0
    DM[1, 1] = 2.0

    dens = psi_values.reduce_orbital_products(DM.tocsr(), DM.lattice)
    predicted = (orb_0**2) * 1.0 + (orb_1**2) * 2.0

    assert np.allclose(dens.grid.ravel(), predicted)

    # Compute the orbital products with an off-diagonal coefficient
    DM = sisl.DensityMatrix(geometry, dim=1, dtype=np.float64)
    DM[0, 0] = 1.0
    DM[0, 1] = 2.0

    dens = psi_values.reduce_orbital_products(DM.tocsr(), DM.lattice)
    predicted = (orb_0**2) * 1.0 + (orb_0 * orb_1) * 2.0

    assert np.allclose(dens.grid.ravel(), predicted)

    # Compute the orbital products with both opposite off-diagonal coefficients
    DM = sisl.DensityMatrix(geometry, dim=1, dtype=np.float64)
    DM[0, 0] = 1.0
    DM[0, 1] = 1.0
    DM[1, 0] = 0.5

    dens = psi_values.reduce_orbital_products(DM.tocsr(), DM.lattice)
    predicted = (orb_0**2) * 1.0 + (orb_0 * orb_1) * 1.5

    assert np.allclose(dens.grid.ravel(), predicted)


def test_orbital_products_periodic(geometry):
    """Very simple tests to see if the orbital products are computed correctly
    when there are periodic conditions.
    """
    geometry = geometry.copy()
    geometry.set_nsc([3, 1, 1])
    geometry.lattice.set_boundary_condition(a="periodic", b="open", c="open")

    # Don't get periodic contributions for orbital values, then predicting
    # grid values will be much easier.
    psi_values = geometry._orbital_values((10, 10, 10))

    orb_csr = psi_values._csr.tocsr()
    orb_0 = orb_csr[:, [0, geometry.no, 2 * geometry.no]].toarray()
    orb_1 = orb_csr[:, [1, geometry.no + 1, 2 * geometry.no + 1]].toarray()

    # Compute the orbital products with one coefficient
    DM = sisl.DensityMatrix(geometry, dim=1, dtype=np.float64)
    DM[0, 0] = 1.0

    dens = psi_values.reduce_orbital_products(DM.tocsr(), geometry.lattice)

    predicted = (orb_0**2).sum(axis=1)

    assert np.allclose(dens.grid.ravel(), predicted)

    # Compute the orbital products with two diagonal coefficients
    DM = sisl.DensityMatrix(geometry, dim=1, dtype=np.float64)
    DM[0, 0] = 1.0
    DM[1, 1] = 2.0

    dens = psi_values.reduce_orbital_products(DM.tocsr(), DM.lattice)
    predicted = (orb_0**2).sum(axis=1) + (orb_1**2).sum(axis=1) * 2.0

    assert np.allclose(dens.grid.ravel(), predicted)

    # Predicting the result with periodic non-diagonal coefficients is a bit more involved.
    # We don't check for now.


@pytest.mark.parametrize("ncoeffs", [1, 2])
def test_orbital_products_onthefly_reduction(geometry, psi_values, ncoeffs):
    """Checks that the on the fly reduction produces the same
    results as computing the whole grid and then reducing."""

    # Build a test DM
    DM = sisl.DensityMatrix(geometry, dim=1, dtype=np.float64)
    DM[0, 0] = 1.0
    DM[1, 1] = 2.0
    DM[2, 2] = 0.75
    DM[3, 3] = 0.92
    DM[0, 1] = 0.5
    DM[1, 0] = 0.5

    csr = DM.tocsr()
    # If this run is to be done with multiple coefficients, build the csr
    # matrix with dim > 1.
    if ncoeffs > 1:
        csr = sisl.SparseCSR(csr, dim=ncoeffs)
        csr.data[:, :] = csr.data[:, 0].reshape(-1, 1)

    # Compute the values on the full grid
    not_reduced = psi_values.reduce_orbital_products(csr, DM.lattice)

    # Reducing the last axis
    reduced = psi_values.reduce_orbital_products(csr, DM.lattice, reduce_grid=(2,))

    post_reduced = not_reduced.sum(2)
    assert reduced.size == post_reduced.size
    if ncoeffs == 1:
        assert np.allclose(reduced.grid, post_reduced.grid, atol=1e-7)
    else:
        assert np.allclose(reduced.reshape(post_reduced.shape), post_reduced)

    # Reducing the two last axes
    reduced = psi_values.reduce_orbital_products(csr, DM.lattice, reduce_grid=(2, 1))

    post_reduced = not_reduced.sum(2).sum(1)
    assert reduced.size == post_reduced.size
    if ncoeffs == 1:
        assert np.allclose(reduced.grid, post_reduced.grid)
    else:
        assert np.allclose(reduced.reshape(post_reduced.shape), post_reduced)

    # Now we are going to reduce the leading axes. This is more involved, so
    # it is more likely to fail.

    # Reducing the first axis
    reduced = psi_values.reduce_orbital_products(csr, DM.lattice, reduce_grid=(0,))

    post_reduced = not_reduced.sum(0)
    assert reduced.size == post_reduced.size
    if ncoeffs == 1:
        assert np.allclose(reduced.grid, post_reduced.grid)
    else:
        assert np.allclose(reduced.reshape(post_reduced.shape), post_reduced)

    # Reducing the first and second axis
    reduced = psi_values.reduce_orbital_products(csr, DM.lattice, reduce_grid=(0, 1))

    post_reduced = not_reduced.sum(1).sum(0)
    assert reduced.size == post_reduced.size
    if ncoeffs == 1:
        assert np.allclose(reduced.grid, post_reduced.grid)
    else:
        assert np.allclose(reduced.reshape(post_reduced.shape), post_reduced)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl import Atom, AtomGhost, Atoms, AtomUnknown, Orbital

pytestmark = [pytest.mark.atom]


@pytest.fixture
def setup():
    class t:
        def __init__(self):
            self.C = Atom["C"]
            self.C3 = Atom("C", [-1] * 3)
            self.Au = Atom("Au")

    return t()


def strings(atom):
    assert isinstance(str(atom), str)
    assert isinstance(repr(atom), str)


def test_atom_simple(setup):
    assert setup.C == Atom["C"]
    assert setup.C == Atom["6"]
    assert setup.C == Atom[setup.C]
    assert setup.C == Atom[Atom["C"]]
    assert setup.C == Atom(Atom["C"])
    assert setup.C == Atom[Atom(6)]
    assert setup.Au == Atom["Au"]
    assert setup.Au != setup.C
    assert setup.Au == setup.Au.copy()


def test_atom_ghost():
    assert isinstance(Atom[1], Atom)
    assert isinstance(Atom[-1], AtomGhost)
    assert isinstance(Atom(1), Atom)
    assert isinstance(Atom(-1), AtomGhost)
    assert isinstance(AtomGhost(-1), AtomGhost)
    assert AtomGhost(-1).Z == 1
    assert Atom(-1).Z == 1


def test_atom_weird():
    for el in (6, "C", "Carbon"):
        atom = Atom(el)
        assert atom.Z == 6
        assert atom.symbol == "C"
        strings(atom)

        atom = Atom[el]
        assert atom.Z == 6
        assert atom.symbol == "C"
        strings(atom)


def test_atom_unknown():
    for el in (1000, "1000"):
        atom = Atom(el)
        assert isinstance(atom, AtomUnknown)
        strings(atom)

        atom = Atom[el]
        assert isinstance(atom, AtomUnknown)
        strings(atom)

    # negative ones are still considered ghosts!
    for el in (-1000, "-1000"):
        atom = Atom(el)
        assert isinstance(atom, AtomGhost)
        strings(atom)

        atom = Atom[el]
        assert isinstance(atom, AtomGhost)
        strings(atom)


def test2(setup):
    C = Atom("C", R=20)
    assert setup.C != C
    Au = Atom("Au", R=20)
    assert setup.C != Au
    C = Atom["C"]
    assert setup.Au != C
    Au = Atom["Au"]
    assert setup.C != Au


def test3(setup):
    assert setup.C.symbol == "C"
    assert setup.C.tag == "C"
    assert setup.Au.symbol == "Au"


def test4(setup):
    assert setup.C.mass > 0
    assert setup.Au.mass > 0
    assert setup.Au.q0 == pytest.approx(0)


def test5(setup):
    assert Atom(Z=1, mass=12).R < 0
    assert Atom(Z=1, mass=12).R.size == 1
    assert Atom(Z=1, mass=12).mass == 12
    assert Atom(Z=31, mass=12).mass == 12
    assert Atom(Z=31, mass=12).Z == 31


def test6(setup):
    assert Atom(1, [-1] * 3).no == 3
    assert len(Atom(1, [-1] * 3)) == 3
    assert Atom(1, 1.4).R == 1.4
    assert Atom(Z=1, R=1.4).R == 1.4
    assert Atom(Z=1, R=1.4).maxR() == 1.4
    assert Atom(Z=1, R=[1.4, 1.8]).no == 2
    assert Atom(Z=1, R=[1.4, 1.8]).maxR() == 1.8
    assert Atom(Z=1, R=[1.4, 1.8]).maxR() == 1.8
    a = Atom(1, Orbital(1.4))
    assert a.R == 1.4
    assert a.no == 1
    a = Atom(1, [Orbital(1.4)])
    assert a.R == 1.4
    assert a.no == 1


def test7(setup):
    assert Atom(1, [-1] * 3).radius() > 0.0
    assert len(str(Atom(1, [-1] * 3)))


def test_fail_equal():
    assert Atom(1.2) != 2.0


def test_tag1():
    a = Atom(6, tag="my-tag")
    assert a.tag == "my-tag"


def test_negative1():
    a = Atom(-1)
    assert a.symbol == "H"
    assert a.tag == "ghost"
    assert a.Z == 1


def test_iter1():
    r = [1, 2]
    a = Atom(5, r)
    for i, o in enumerate(a):
        assert o.R == r[i]
    for i, o in enumerate(a.iter()):
        assert o.R == r[i]


def test_iter2():
    r = [1, 1, 2, 2]
    a = Atom(5, r)
    for i, o in enumerate(a.iter(True)):
        assert len(o) == 2


def test_charge():
    r = [1, 1, 2, 2]
    a = Atom(5, [Orbital(1.0, 1.0), Orbital(1.0, 1.0), Orbital(2.0), Orbital(2.0)])
    assert len(a.q0) == 4
    assert a.q0.sum() == pytest.approx(2)


def test_atoms_set():
    a1 = Atom(1)
    a2 = Atom(2)
    a3 = Atom(3)
    a = Atoms(a1, 3)
    assert len(a) == 3
    assert len(a.atom) == 1
    a[1] = a2
    assert len(a) == 3
    assert len(a.atom) == 2

    # Add the atom, but do not specify any
    # atoms to have the species
    a[[]] = a3
    assert len(a) == 3
    assert len(a.atom) == 3
    for atom in a:
        assert atom in [a1, a2]
        assert atom != a3
    found = 0
    for atom, i_s in a.iter(True):
        if atom == a1:
            assert len(i_s) == 2
            found += 1
        elif atom == a2:
            assert len(i_s) == 1
            found += 1
        elif atom == a3:
            assert len(i_s) == 0
            found += 1
    assert found == 3


def test_charge_diff():
    o1 = Orbital(1.0, 1.0)
    o2 = Orbital(1.0, 0.5)
    a1 = Atom(5, [o1, o2, o1, o2])
    a2 = Atom(5, [o1, o2, o1, o2, o1, o1])
    assert len(a1.q0) == 4
    assert len(a2.q0) == 6
    assert a1.q0.sum() == pytest.approx(3)
    assert a2.q0.sum() == pytest.approx(5)


def test_multiple_orbitals():
    o = [Orbital(1.0, 1.0), Orbital(2.0, 0.5), Orbital(3.0, 0.75)]
    a1 = Atom(5, o)
    assert len(a1) == 3
    for i in range(3):
        a2 = a1.sub([i])
        assert len(a2) == 1
        assert a2.orbitals[0] == o[i]
        assert a2.orbitals[0] == a2[0]
        a2 = a1.remove([i])
        assert len(a2) == 2

    a2 = a1.sub([1, 2, 0])
    assert a2.orbitals[0] == o[1]
    assert a2[0] == o[1]
    assert a2.orbitals[1] == o[2]
    assert a2[1] == o[2]
    assert a2.orbitals[2] == o[0]
    assert a2[2] == o[0]


def test_multiple_orbitals_fail_io():
    o = [Orbital(1.0, 1.0), Orbital(2.0, 0.5), Orbital(3.0, 0.75)]
    with pytest.raises(ValueError):
        Atom(5, o).sub([3])


def test_multiple_orbitals_fail_len():
    o = [Orbital(1.0, 1.0), Orbital(2.0, 0.5), Orbital(3.0, 0.75)]
    with pytest.raises(ValueError):
        Atom(5, o).sub([0, 0, 0, 0, 0])


def test_atom_getattr_orbs():
    class UOrbital(Orbital):
        def __init__(self, *args, **kwargs):
            U = kwargs.pop("U", 0.0)
            super().__init__(*args, **kwargs)
            self._U = U

        @property
        def U(self):
            return self._U

    o = [
        UOrbital(1.0, 1.0, U=1.0),
        UOrbital(2.0, 0.5, U=2.0),
        UOrbital(3.0, 0.75, U=3.0),
    ]
    a = Atom(5, o)
    assert np.allclose(a.U, [1, 2, 3])
    # also test the callable interface
    assert all(map(lambda x: len(x) == 0, a.name()))


def test_atoms_init_tags():
    a = Atoms([dict(Z="H", tag="hello"), "B", "H"])
    assert len(a) == 3
    assert len(a.atom) == 3
    assert a.atom[0].tag == "hello"
    assert a.atom[1].tag == "B"
    assert a.atom[2].tag == "H"

    a = Atoms(dict(Z="H", tag="hello"))
    assert len(a) == 1
    assert len(a.atom) == 1
    assert a.atom[0].tag == "hello"


def test_atoms_formula():
    a = Atoms([dict(Z="H", tag="hello"), "B", "H"])
    assert len(a) == 3
    assert len(a.atom) == 3
    assert a.formula() == "BH2"


def test_atoms_formula_unknown_system():
    a = Atoms([dict(Z="H", tag="hello"), "B", "H"])
    with pytest.raises(ValueError):
        assert a.formula("not_known")


def test_atom_orbitals():
    a = Atom(5, ["x", "y"])
    assert len(a) == 2
    assert a[0].tag == "x"
    assert a[1].tag == "y"
    assert a.maxR() < 0.0
    a = Atom(5, [1.2, 1.4])
    assert len(a) == 2
    assert a.maxR() == pytest.approx(1.4)


def test_pickle(setup):
    import pickle as p

    sC = p.dumps(setup.C)
    sC3 = p.dumps(setup.C3)
    sAu = p.dumps(setup.Au)
    C = p.loads(sC)
    C3 = p.loads(sC3)
    Au = p.loads(sAu)
    assert Au != C
    assert C == setup.C
    assert C3 == setup.C3
    assert C != setup.Au
    assert Au == setup.Au
    assert Au != setup.C



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl import Atom, Atoms
from sisl.io.siesta.orb_indx import *

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def test_si_pdos_kgrid_orb_indx(sisl_files):
    f = sisl_files("siesta", "Si_pdos_k", "Si_pdos.ORB_INDX")
    nsc = orbindxSileSiesta(f).read_lattice_nsc()
    assert np.all(nsc > 1)
    atoms = orbindxSileSiesta(f).read_basis()

    assert len(atoms) == 2
    assert atoms.nspecies == 1
    assert atoms.atom[0].Z == 14
    assert len(atoms[0]) == 9
    assert len(atoms[1]) == 9


def test_h2o_dipole_orb_indx(sisl_files):
    f = sisl_files("siesta", "H2O_dipole", "h2o_dipole.ORB_INDX")

    nsc = orbindxSileSiesta(f).read_lattice_nsc()
    assert np.all(nsc == 1)
    atoms = orbindxSileSiesta(f).read_basis()

    assert len(atoms) == 6
    assert atoms.nspecies == 2
    Z = np.zeros(len(atoms))
    Z[:] = 1
    Z[[0, 3]] = 8

    assert np.allclose(atoms.Z, Z)
    # number of orbitals on each atom
    assert len(atoms[0]) == 9
    assert len(atoms[-1]) == 4


def test_orb_indx_order(sisl_tmp):
    f = sisl_tmp("test.ORD_INDX")

    with open(f, "w") as fh:
        fh.write(
            """\
     29  29 = orbitals in unit cell and supercell. See end of file.

    io    ia is   spec iao  n  l  m  z  p          sym      rc    isc     iuo
     1     1  2 Fe_SOC   1  4  0  0  1  F            s   7.329  0  0  0     1
     2     1  2 Fe_SOC   2  4  0  0  2  F            s   6.153  0  0  0     2
     3     1  2 Fe_SOC   3  3  2 -2  1  F          dxy   4.336  0  0  0     3
     4     1  2 Fe_SOC   4  3  2 -1  1  F          dyz   4.336  0  0  0     4
     5     1  2 Fe_SOC   5  3  2  0  1  F          dz2   4.336  0  0  0     5
     6     1  2 Fe_SOC   6  3  2  1  1  F          dxz   4.336  0  0  0     6
     7     1  2 Fe_SOC   7  3  2  2  1  F       dx2-y2   4.336  0  0  0     7
     8     1  2 Fe_SOC   8  3  2 -2  2  F          dxy   2.207  0  0  0     8
     9     1  2 Fe_SOC   9  3  2 -1  2  F          dyz   2.207  0  0  0     9
    10     1  2 Fe_SOC  10  3  2  0  2  F          dz2   2.207  0  0  0    10
    11     1  2 Fe_SOC  11  3  2  1  2  F          dxz   2.207  0  0  0    11
    12     1  2 Fe_SOC  12  3  2  2  2  F       dx2-y2   2.207  0  0  0    12
    13     1  2 Fe_SOC  13  4  1 -1  1  T          Ppy   7.329  0  0  0    13
    14     1  2 Fe_SOC  14  4  1  0  1  T          Ppz   7.329  0  0  0    14
    15     1  2 Fe_SOC  15  4  1  1  1  T          Ppx   7.329  0  0  0    15
    16     2  1     Pt   1  6  0  0  1  F            s   7.158  0  0  0    16
    17     2  1     Pt   2  6  0  0  2  F            s   6.009  0  0  0    17
    18     2  1     Pt   3  5  2 -2  1  F          dxy   5.044  0  0  0    18
    19     2  1     Pt   4  5  2 -1  1  F          dyz   5.044  0  0  0    19
    20     2  1     Pt   5  5  2  0  1  F          dz2   5.044  0  0  0    20
    21     2  1     Pt   6  5  2  1  1  F          dxz   5.044  0  0  0    21
    22     2  1     Pt   7  5  2  2  1  F       dx2-y2   5.044  0  0  0    22
    23     2  1     Pt   8  5  2 -2  2  F          dxy   3.022  0  0  0    23
    24     2  1     Pt   9  5  2 -1  2  F          dyz   3.022  0  0  0    24
    25     2  1     Pt  10  5  2  0  2  F          dz2   3.022  0  0  0    25
    26     2  1     Pt  11  5  2  1  2  F          dxz   3.022  0  0  0    26
    27     2  1     Pt  12  5  2  2  2  F       dx2-y2   3.022  0  0  0    27
    28     2  1     Pt  13  6  1 -1  1  T          Ppy   7.158  0  0  0    28
    29     2  1     Pt  14  6  1  0  1  T          Ppz   7.158  0  0  0    29
"""
        )

    atoms = orbindxSileSiesta(f).read_basis()

    assert len(atoms) == 2
    assert atoms.nspecies == 2
    Z = [1001, 78]
    assert np.allclose(atoms.Z, Z)
    # number of orbitals on each atom
    assert len(atoms[0]) == 15
    assert len(atoms[1]) == 14
    assert atoms.atom[0].tag == "Pt"
    assert len(atoms.atom[0]) == 14
    assert atoms.atom[1].tag == "Fe_SOC"
    assert len(atoms.atom[1]) == 15

    atoms = orbindxSileSiesta(f).read_basis(atoms=Atoms([Atom("Pt"), Atom("Fe")]))

    assert len(atoms) == 2
    assert atoms.nspecies == 2
    Z = [26, 78]
    assert np.allclose(atoms.Z, Z)
    # number of orbitals on each atom
    assert len(atoms[0]) == 15
    assert len(atoms[1]) == 14
    assert atoms.atom[0].tag == "Pt"
    assert len(atoms.atom[0]) == 14
    assert atoms.atom[1].tag == "Fe_SOC"
    assert len(atoms.atom[1]) == 15



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl import Atom, Atoms, AtomUnknown, Geometry
from sisl.io.siesta.xv import *

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def test_xv1(sisl_tmp, sisl_system):
    f = sisl_tmp("gr.XV")
    sisl_system.g.write(xvSileSiesta(f, "w"))
    g = xvSileSiesta(f).read_geometry()

    # Assert they are the same
    assert np.allclose(g.cell, sisl_system.g.cell)
    assert np.allclose(g.xyz, sisl_system.g.xyz)
    assert sisl_system.g.atoms.equal(g.atoms, R=False)


def test_xv_reorder(sisl_tmp, sisl_system):
    f = sisl_tmp("gr.XV")
    g = sisl_system.g.copy()
    g.atoms[0] = Atom(1)
    g.write(xvSileSiesta(f, "w"))
    g2 = xvSileSiesta(f).read_geometry()

    # Assert they are the same
    assert np.allclose(g.cell, g2.cell)
    assert np.allclose(g.xyz, g2.xyz)
    assert g.atoms.equal(g2.atoms, R=False)


def test_xv_velocity(sisl_tmp, sisl_system):
    f = sisl_tmp("gr.XV")
    g = sisl_system.g.copy()
    g.atoms[0] = Atom(1)
    v = np.random.rand(len(g), 3)
    g.write(xvSileSiesta(f, "w"), velocity=v)

    # Try to read in different ways
    g2 = xvSileSiesta(f).read_geometry()
    assert np.allclose(g.cell, g2.cell)
    assert np.allclose(g.xyz, g2.xyz)
    assert g.atoms.equal(g2.atoms, R=False)

    g2, v2 = xvSileSiesta(f).read_geometry(ret_velocity=True)
    assert np.allclose(g.cell, g2.cell)
    assert np.allclose(g.xyz, g2.xyz)
    assert g.atoms.equal(g2.atoms, R=False)
    assert np.allclose(v, v2)

    # Try to read in different ways
    v2 = xvSileSiesta(f).read_velocity()
    assert np.allclose(v, v2)


def test_xv_ghost(sisl_tmp):
    f = sisl_tmp("ghost.XV")
    a1 = Atom(1)
    am1 = Atom(-1)
    g = Geometry([[0.0, 0.0, i] for i in range(2)], [a1, am1], 2.0)
    g.write(xvSileSiesta(f, "w"))

    g2 = xvSileSiesta(f).read_geometry()
    assert np.allclose(g.cell, g2.cell)
    assert np.allclose(g.xyz, g2.xyz)
    assert np.allclose(g.atoms.Z, g2.atoms.Z)
    assert g.atoms[0].__class__ is g2.atoms[0].__class__
    assert g.atoms[1].__class__ is g2.atoms[1].__class__
    assert g.atoms[0].__class__ is not g2.atoms[1].__class__


def test_xv_missing_atoms(sisl_tmp):
    # test for #778
    f = sisl_tmp("missing.XV")
    with open(f, "w") as fh:
        fh.write(
            """\
1. 0. 0.  0. 0. 0.
0. 1. 0.  0. 0. 0.
0. 0. 2.  0. 0. 0.
6
2 6 0. 1. 0.  0. 0. 0.
1 2 0. 1. 0.  0. 0. 0.
4 3 0. 1. 0.  0. 0. 0.
4 3 0. 1. 0.  0. 0. 0.
1 2 0. 1. 0.  0. 0. 0.
2 6 0. 1. 0.  0. 0. 0.
"""
        )
    geom = xvSileSiesta(f).read_geometry()
    assert len(geom) == 6
    assert len(geom.atoms.atom) == 4
    assert np.allclose(geom.atoms.species, [1, 0, 3, 3, 0, 1])

    # start_Z + sp_idx
    atom = AtomUnknown(1000 + 2)
    assert geom.atoms.atom[2].Z == atom.Z
    assert isinstance(geom.atoms.atom[2], atom.__class__)


def test_xv_missing_atoms_end(sisl_tmp):
    # test for #778
    f = sisl_tmp("missing_end.XV")
    with open(f, "w") as fh:
        fh.write(
            """\
1. 0. 0.  0. 0. 0.
0. 1. 0.  0. 0. 0.
0. 0. 2.  0. 0. 0.
6
2 6 0. 1. 0.  0. 0. 0.
1 2 0. 1. 0.  0. 0. 0.
1 2 0. 1. 0.  0. 0. 0.
3 3 0. 1. 0.  0. 0. 0.
3 3 0. 1. 0.  0. 0. 0.
2 6 0. 1. 0.  0. 0. 0.
"""
        )
    atoms = Atoms([2, 6, 3, 5])
    geom = xvSileSiesta(f).read_geometry(atoms=atoms)
    assert len(geom) == 6
    assert len(geom.atoms.atom) == 4
    assert np.allclose(geom.atoms.species, [1, 0, 0, 2, 2, 1])


def test_xv_missing_atoms_species(sisl_tmp):
    # test for #778
    f = sisl_tmp("missing_species.XV")
    with open(f, "w") as fh:
        fh.write(
            """\
1. 0. 0.  0. 0. 0.
0. 1. 0.  0. 0. 0.
0. 0. 2.  0. 0. 0.
6
2 1 0. 1. 0.  0. 0. 0.
1 1 0. 1. 0.  0. 0. 0.
1 1 0. 1. 0.  0. 0. 0.
3 1 0. 1. 0.  0. 0. 0.
3 1 0. 1. 0.  0. 0. 0.
2 1 0. 1. 0.  0. 0. 0.
"""
        )

    atoms = Atoms([Atom(1, tag=tag) for tag in "ABCD"])
    geom = xvSileSiesta(f).read_geometry(atoms=atoms)
    assert len(geom) == 6
    assert len(geom.atoms.atom) == 4
    assert np.allclose(geom.atoms.species, [1, 0, 0, 2, 2, 1])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl.io.xyz import *

pytestmark = [pytest.mark.io, pytest.mark.generic]


def test_xyz1(sisl_tmp, sisl_system):
    f = sisl_tmp("gr.xyz")
    sisl_system.g.write(xyzSile(f, "w"))
    g = xyzSile(f).read_geometry()

    # Assert they are the same
    assert np.allclose(g.cell, sisl_system.g.cell)
    assert np.allclose(g.xyz, sisl_system.g.xyz)
    assert sisl_system.g.atoms.equal(g.atoms, R=False)
    g = xyzSile(f).read_geometry(lattice=g.lattice)
    assert np.allclose(g.cell, sisl_system.g.cell)


def test_xyz_sisl(sisl_tmp):
    f = sisl_tmp("sisl.xyz")

    with open(f, "w") as fh:
        fh.write(
            """3
sisl-version=1 nsc=1 1 3 cell=10 0 0 0 12 0 0 0 13
C   0.00000000  0.00000000  0.00000000
C   1.000000  0.00000000  0.00000000
C   2.00000  0.00000000  0.00000000
"""
        )
    g = xyzSile(f).read_geometry()

    # Assert they are the same
    assert np.allclose(g.cell, [[10, 0, 0], [0, 12, 0], [0, 0, 13]])
    assert np.allclose(g.xyz[:, 0], [0, 1, 2])
    assert np.allclose(g.xyz[:, 1], 0.0)
    assert np.allclose(g.xyz[:, 2], 0.0)
    assert np.allclose(g.nsc, [1, 1, 3])

    g = xyzSile(f).read_geometry(lattice=[10, 11, 13])
    assert np.allclose(g.cell, [[10, 0, 0], [0, 11, 0], [0, 0, 13]])


def test_xyz_ase(sisl_tmp):
    f = sisl_tmp("ase.xyz")
    with open(f, "w") as fh:
        fh.write(
            """3
Lattice="10 0 0 0 12 0 0 0 13" Properties=species:S:1:pos:R:3 pbc="F F T"
C   0.00000000  0.00000000  0.00000000
C   1.000000  0.00000000  0.00000000
C   2.00000  0.00000000  0.00000000
"""
        )
    g = xyzSile(f).read_geometry()

    # Assert they are the same
    assert np.allclose(g.cell, [[10, 0, 0], [0, 12, 0], [0, 0, 13]])
    assert np.allclose(g.xyz[:, 0], [0, 1, 2])
    assert np.allclose(g.xyz[:, 1], 0.0)
    assert np.allclose(g.xyz[:, 2], 0.0)
    assert np.allclose(g.nsc, [1, 1, 1])
    assert np.allclose(g.pbc, [False, False, True])


def test_xyz_arbitrary(sisl_tmp):
    f = sisl_tmp("ase.xyz")
    with open(f, "w") as fh:
        fh.write(
            """3

C   0.00000000  0.00000000  0.00000000
C   1.000000  0.00000000  0.00000000
C   2.00000  0.00000000  0.00000000
"""
        )
    g = xyzSile(f).read_geometry()

    # Assert they are the same
    assert np.allclose(g.xyz[:, 0], [0, 1, 2])
    assert np.allclose(g.xyz[:, 1], 0.0)
    assert np.allclose(g.xyz[:, 2], 0.0)
    assert np.allclose(g.nsc, [1, 1, 1])


def test_xyz_multiple(sisl_tmp):
    f = sisl_tmp("sisl_multiple.xyz")
    with open(f, "w") as fh:
        fh.write(
            """1

C   0.00000000  0.00000000  0.00000000
2

C   0.00000000  0.00000000  0.00000000
C   1.000000  0.00000000  0.00000000
3

C   0.00000000  0.00000000  0.00000000
C   1.000000  0.00000000  0.00000000
C   2.00000  0.00000000  0.00000000
"""
        )
    g = xyzSile(f).read_geometry()
    assert g.na == 1
    g = xyzSile(f).read_geometry[1]()
    assert g.na == 2
    g = xyzSile(f).read_geometry[:]()
    assert len(g) == 3
    assert g[0].na == 1 and g[1].na == 2 and g[-1].na == 3
    g = xyzSile(f).read_geometry[1:]()
    assert len(g) == 2
    assert g[0].na == 2 and g[-1].na == 3
    g = xyzSile(f).read_geometry[1:-1]()
    assert len(g) == 1
    assert g[0].na == 2
    g = xyzSile(f).read_geometry[::2]()
    assert len(g) == 2
    assert g[0].na == 1 and g[-1].na == 3
    g = xyzSile(f).read_geometry[:2:1]()
    assert len(g) == 2
    assert g[0].na == 1 and g[1].na == 2
    g = xyzSile(f).read_geometry[1:3:1]()
    assert len(g) == 2
    assert g[0].na == 2 and g[1].na == 3

    g1 = xyzSile(f).read_geometry[-1]()
    g2 = xyzSile(f).read_geometry.last()
    assert g1 == g2

    # ensure it works with other arguments
    g = xyzSile(f).read_geometry(lattice=None, atoms=None)
